/*
**  MailboxManagerController.m
**
**  Copyright (C) 2001-2004
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "MailboxManagerController.h"

#include "ConsoleWindowController.h"
#include "Constants.h"
#include "EditWindowController.h"
#include "ExtendedOutlineView.h"
#include "Filter.h"
#include "FilterManager.h"
#include "GNUMail.h"
#include "Task.h"
#include "TaskManager.h"

#ifndef MACOSX
#include "ImageTextCell.h"
#include "MailboxManager.h"
#endif

#include "FolderNode.h"
#include "MailboxManagerCache.h"
#include "MailWindowController.h"
#include "MessageViewWindowController.h"
#include "NewMailboxPanelController.h"
#include "NSUserDefaults+Extensions.h"
#include "Utilities.h"

#include <Pantomime/Constants.h>
#include <Pantomime/Flags.h>
#include <Pantomime/FolderInformation.h>
#include <Pantomime/IMAPCacheManager.h>
#include <Pantomime/IMAPFolder.h>
#include <Pantomime/IMAPStore.h>
#include <Pantomime/LocalFolder.h>
#include <Pantomime/LocalStore.h>
#include <Pantomime/Message.h>
#include <Pantomime/NSData+Extensions.h>
#include <Pantomime/NSString+Extensions.h>
#include <Pantomime/TCPConnection.h>
#include <Pantomime/URLName.h>
#include <Pantomime/VirtualFolder.h>

#define UPDATE_PATH(name, theOldPath, thePath) ({ \
 if ( [[allValues objectForKey: name] isEqualToString: theOldPath] ) \
   { \
     [allValues setObject: thePath  forKey: name]; \
   } \
})

static MailboxManagerController *singleInstance = nil;

// 
// Here's how it does work:
//
// _allFolders (NSArray) -> localNodes (FolderNode * - name == _("Local"))
// 
//                       -> IMAP FolderNode 1 (FoderNode * - name == "username @ imap.server1.com")  
//                            
//                       -> IMAP FolderNode 2 (FoderNode * - name == "username @ imap.server2.com")
//  
//                       -> ...   
//                                 
//
@implementation MailboxManagerController

#ifndef MACOSX
- (id) initWithWindowNibName: (NSString *) windowNibName
{
  NSToolbar *aToolbar;
  id aCell;

  MailboxManager *theWindow;
  
  theWindow = [[MailboxManager alloc] initWithContentRect: NSMakeRect(200,200,246,435)
				      styleMask: NSClosableWindowMask|NSTitledWindowMask|NSMiniaturizableWindowMask|NSResizableWindowMask
				      backing: NSBackingStoreBuffered
				      defer: YES];

  self = [super initWithWindow: theWindow];

  [theWindow layoutWindow];
  [theWindow setDelegate: self];

  // We link our outlets
  outlineView = theWindow->outlineView;
  scrollView = theWindow->scrollView;
  RELEASE(theWindow);

  // We set the title of our window (causing it to be loaded under OS X)
  [[self window] setTitle: _(@"Mailboxes")];

  aToolbar = [[NSToolbar alloc] initWithIdentifier: @"MailboxManagerToolbar"];
  [aToolbar setDelegate: self];
  [aToolbar setAllowsUserCustomization: YES];
  [aToolbar setAutosavesConfiguration: YES];
  [[self window] setToolbar: aToolbar];
  RELEASE(aToolbar);

  // We now set our data cell for the "Mailbox" column
  aCell =  [[ImageTextCell alloc] init];
  [[outlineView tableColumnWithIdentifier: @"Mailbox"] setDataCell: aCell];
  AUTORELEASE(aCell);

  // We register the outline view for dragged types
  [outlineView registerForDraggedTypes: [NSArray arrayWithObject: MessagePboardType]];

  // We set our autosave window frame name and restore the one from the user's defaults.
  [[self window] setFrameAutosaveName: @"MailboxManager"];
  [[self window] setFrameUsingName: @"MailboxManager"]; 

  // We set our autosave name for our outline view
  [outlineView setAutosaveName: @"MailboxManager"];
  [outlineView setAutosaveTableColumns: YES];

  // We set our outline view background color
  if ( [[NSUserDefaults standardUserDefaults] colorForKey: @"MAILBOXMANAGER_OUTLINE_COLOR"] )
    {
      [outlineView setBackgroundColor: [[NSUserDefaults standardUserDefaults]
					 colorForKey: @"MAILBOXMANAGER_OUTLINE_COLOR"]];
      [scrollView setBackgroundColor: [[NSUserDefaults standardUserDefaults]
					colorForKey: @"MAILBOXMANAGER_OUTLINE_COLOR"]];
    }

  return self;
}
#else
- (id) init
{
  self = [super init];
  
  // We initialize some ivars
  [self windowDidLoad];
  
  return self;
}
#endif


//
//
//
- (void) dealloc
{
  NSDebugLog(@"MailboxManagerController: -dealloc");

  [[NSNotificationCenter defaultCenter] removeObserver: self];
  
#ifndef MACOSX
  [[self window] setDelegate: nil];
#endif

  RELEASE(menu);
  RELEASE(localNodes);
  RELEASE(mailboxManagerCache);
  RELEASE(_allFolders);
  RELEASE(allStores);
  
  [super dealloc];
}


//
// Datasource methods for the outline view
//
- (id) outlineView: (NSOutlineView *) outlineView
	     child: (int) index
	    ofItem: (id) item
{
  if ( !item || item == _allFolders )
    {
      return [_allFolders objectAtIndex: index];
    }

  if ( [item isKindOfClass: [FolderNode class]] )
    {
      return [(FolderNode *)item childAtIndex: index];
    }
 
  return nil;
}


//
//
//
- (BOOL) outlineView: (NSOutlineView *) outlineView
    isItemExpandable: (id) item
{
  if ( item == _allFolders || [_allFolders containsObject: item] )
    {
      return YES;
    }
  
  if ( [item isKindOfClass: [FolderNode class]] )
    {
      if ( [(FolderNode *)item childCount] > 0 )
	{
	  return YES;
	}
      else
	{
	  return NO;
	}
    }

  return NO;
}


//
//
//
- (int)        outlineView: (NSOutlineView *) outlineView 
    numberOfChildrenOfItem: (id) item
{
  // The root.
  if (!item || item == _allFolders)
    {
      return [_allFolders count];
    }
  
  // Children of our root, the Local folder and all the IMAP folders, subfolders, etc.
  if ([item isKindOfClass: [FolderNode class]])
    {
      return [(FolderNode *)item childCount];
    }
  
  return 0;
}


//
//
//
- (id)         outlineView: (NSOutlineView *) outlineView 
 objectValueForTableColumn: (NSTableColumn *) tableColumn 
		    byItem: (id) item
{
  if ( [[[tableColumn headerCell] stringValue] isEqual: _(@"Mailbox")] )
    {
      if ( [item isKindOfClass: [FolderNode class]] )
	{
	  return [(FolderNode *)item name];
	}
    }
  
  if ( [item isKindOfClass: [FolderNode class]] && [item parent] )
    {
      BOOL b;
      
      b = [[Utilities completePathForFolderNode: item  separator: @"/"]
	    hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]];
      
      if ( [item childCount] == 0 || !b )
	{
	  int nbOfMessages, nbOfUnreadMessages;
	  
	  [self _nbOfMessages: &nbOfMessages
		nbOfUnreadMessages: &nbOfUnreadMessages
		forItem: item];
	  
	  // If we have an IMAP folder AND the could is 0 AND it has children, do nothing.
	  if (!b && nbOfMessages == 0 && [item childCount] > 0) return nil;

	  if ( [[[tableColumn headerCell] stringValue] isEqual: _(@"Messages")] )
	    {
	      return [NSString stringWithFormat: _(@"%d (%d unread)"), nbOfMessages, nbOfUnreadMessages];
	    }
	}
    }
  
  return nil;
}


//
//
//
- (void) outlineView: (NSOutlineView *) theOutlineView
      setObjectValue: (id) theObject
      forTableColumn: (NSTableColumn *) theTableColumn
	      byItem: (id) theItem
{
  // If the previous name is different than the new one, we proceed.
  if (![[(FolderNode *)theItem name] isEqualToString: theObject])
    {
      NSString *pathOfFolder;
      Store *aStore;

      aStore = [self storeForFolderNode: theItem];

      //
      // pathOfFolder will hold a value like:  folderA
      //                                       folderA/folderB
      //                                       folderA/folderB/folderC
      //                                  or:
      //                                       folderA.folderB
      //                                       folderA.folderB.folderC
      //
      pathOfFolder = [Utilities pathOfFolderFromFolderNode: theItem
				separator: [(id<Store>)aStore folderSeparator]];
      
      [(id<Store>)aStore renameFolderWithName: [pathOfFolder stringByDeletingFirstPathSeparator: [(id<Store>)aStore folderSeparator]]
		  toName: [[NSString stringWithFormat: @"%@%@%@",
			      [pathOfFolder stringByDeletingLastPathComponentWithSeparator: [(id<Store>)aStore folderSeparator]],
				     [(id<Store>)aStore folderSeparator], theObject] 
			    stringByDeletingFirstPathSeparator: [(id<Store>)aStore folderSeparator]]];
    }
}


//
// Delegate method used to prevent the user from renaming
// "invalid" mailboxes / folders.
//
- (BOOL)   outlineView: (NSOutlineView *) theOutlineView
 shouldEditTableColumn: (NSTableColumn *) theTableColumn
		  item: (id) theItem
{
  int row, level;
  id item;

  row = [theOutlineView selectedRow];

  if ( row < 0 )
    {
      return NO;
    }
  
  item = [theOutlineView itemAtRow: row];
  level = [theOutlineView levelForItem: item];
  
  if ([theOutlineView numberOfSelectedRows] != 1 || level < 1)
    {
      return NO;
    }

  return YES;
}


//
//
//
- (void) outlineView: (NSOutlineView *) aOutlineView
     willDisplayCell: (id) aCell
      forTableColumn: (NSTableColumn *) aTableColumn
                item: (id) item
{
  // We set our default node icon, if we need to.
  if ([[[aTableColumn headerCell] stringValue] isEqual: _(@"Mailbox")])
    {
      int level;
      
      level = [aOutlineView levelForItem: item];
      
      if ( level > 0 && [(FolderNode *)item childCount] == 0 )
	{
	  NSString *aString;
	  id aStore;
	  
	  aString = [self _stringValueOfURLNameFromItem: item
			  store: &aStore];

	  if ( [Utilities stringValueOfURLName: aString  isEqualTo: @"TRASHFOLDERNAME"] )
	    {
	      [aCell setImage: [NSImage imageNamed: @"delete_16"]];
	    }
	  else if ( [Utilities stringValueOfURLName: aString  isEqualTo: @"SENTFOLDERNAME"] )
	    {
	      [aCell setImage: [NSImage imageNamed: @"send_16"]];
	    }
	  else if ( [Utilities stringValueOfURLName: aString  isEqualTo: @"DRAFTSFOLDERNAME"] )
	    {
	      [aCell setImage: [NSImage imageNamed: @"create_16"]];
	    }
	  else if ( [Utilities stringValueOfURLName: aString  isEqualTo: @"INBOXFOLDERNAME"] )
	    { 
	      [aCell setImage: [NSImage imageNamed: @"retrieve_16"]];
	    }
	  else
	    {
#ifdef MACOSX
	      [aCell setImage: [NSImage imageNamed: @"openFolder_osx"]];
#else
	      [aCell setImage: [NSImage imageNamed: @"openFolder"]];
#endif
	    }
	} 
      else 
	{ 
	  [aCell setImage: nil];
	}
    }

  //
  //
  //
  if ( [item isKindOfClass: [FolderNode class]] && [item parent] )
    {
      int nbOfMessages, nbOfUnreadMessages;
      
      [self _nbOfMessages: &nbOfMessages
	    nbOfUnreadMessages: &nbOfUnreadMessages
	    forItem: item];

      if ( nbOfUnreadMessages > 0 )
	{
#ifdef MACOSX
	  [aCell setFont: [NSFont boldSystemFontOfSize: [NSFont smallSystemFontSize]]];
#else
	  [aCell setFont: [NSFont boldSystemFontOfSize: 0]];
#endif
	  return;
	}
    }
  
  // We set our default font.
#ifdef MACOSX
  [aCell setFont: [NSFont systemFontOfSize: [NSFont smallSystemFontSize]]];
#else
  [aCell setFont: [NSFont systemFontOfSize: 0]];
#endif

  // We set the right text aligment
  if ( [[[aTableColumn headerCell] stringValue] isEqual: _(@"Mailbox")] )
    {
      [aCell setAlignment: NSLeftTextAlignment];
    }
  else
    {
      [aCell setAlignment: NSRightTextAlignment];
    }   
}


//
//
//
- (void) outlineViewSelectionDidChange: (NSNotification *) theNotification
{
  [self open: [theNotification object]];
}


//
//
//
- (NSMenu *) outlineView: (NSOutlineView *) aOutlineView
      contextMenuForItem: (id) item
{
  int row, level;
  BOOL aBOOL;
  
  row = [aOutlineView selectedRow];
  level = [aOutlineView levelForItem: item];
  
  aBOOL = (row > 0 && [aOutlineView numberOfSelectedRows] == 1 && level >= 1);
  
  [[menu itemWithTitle: _(@"Delete")] setEnabled: aBOOL];
  [[menu itemWithTitle: _(@"Rename")] setEnabled: aBOOL];
  
  aBOOL = (row >= 0 && [aOutlineView numberOfSelectedRows] == 1 && level >= 0);
  
  [[menu itemWithTitle: _(@"Create")] setEnabled: aBOOL];

  [menu update];
  return menu;
}


//
// Delegate methods
//
#warning FIXME remove or try to find a better way to disconnect from the store
#if 0
- (BOOL) outlineView: (NSOutlineView *) outlineView
  shouldCollapseItem: (id) item
{
  if ([_allFolders containsObject: item] &&
      item != localNodes)
    {
      NSString *aUsername, *aServerName;
      IMAPStore *aStore;

      BOOL shouldCloseStore;

      [Utilities storeKeyForFolderNode: item
		 serverName: &aServerName
		 username: &aUsername];
      
      aStore = [self storeForName: aServerName
		     username: aUsername];
      shouldCloseStore = NO;
      
      if ([aStore isConnected])
	{
	  int choice;
	  
	  choice = NSRunAlertPanel( [NSString stringWithFormat: _(@"Closing connection on %@ @ %@..."), aUsername, aServerName],
				    _(@"Would you like to close the connection with this IMAP server?"),
				    _(@"Close"), // default
				    _(@"No"),    // alternate
				    NULL );
	  
	  if ( choice == NSAlertDefaultReturn )
	    {
	      shouldCloseStore = YES;
	    }
	}
      
      //
      // The store is already disconnected or the user wants to close it.
      // Let's close its window and remove it from the list of opened stores.
      //
      if (![aStore isConnected] || shouldCloseStore)
	{
	  id aWindow;
	  
	  // We MUST "flush" the folder associated to the window if we are re-using it since
	  // it's currently displaying an IMAP folder of this store (if it hasn't been closed).
	  // We do that ONLY if the store of the displayed folder IS this store.
	  // For example, one could have 'expanded' the IMAP node but not opened any 
	  // IMAP folder so we must NOT automatically "flush" the folder associated with the 
	  // MailWindowController since it could be displaying a local folder!
	  aWindow = [Utilities windowForFolderName: nil  store: (Store *)aStore];
	  
	  if (aWindow)
	    {
	      // Flush the folder...
	      [[aWindow windowController] setFolder: nil];
	    }
	  
	  // This does nothing if it's already disconnected...
	  [aStore close];
	  
	  // This causes problems under OS X. Anyway, the children will be released
	  // if we re-open the connection to the IMAP server.
#ifndef MACOSX	      
	  [item setChildren: nil];
#endif
	  [allStores removeObjectForKey: [Utilities storeKeyForFolderNode: item
						    serverName: NULL
						    username: NULL]];
	  
	  ADD_CONSOLE_MESSAGE(_(@"Closing IMAP connection on %@."), aServerName);
	}
      
    }

  return YES;
}
#endif

//
//
//
- (BOOL) outlineView: (NSOutlineView *) outlineView
    shouldExpandItem: (id) item
{
  if ( item == _allFolders || item == localNodes )
    {
      return YES;
    }
  
  if ( [_allFolders containsObject: item] )
    {
      NSString *aString;

      aString = [(FolderNode *)item name];    // returns the account name

      return [self _initializeIMAPStoreWithAccountName: aString];
    }
  
  return YES;
}


//
// NSOutlineViewDataSource Drag and drop
//
- (NSDragOperation) outlineView: (NSOutlineView*) theOutlineView
		   validateDrop: (id <NSDraggingInfo>) info
		   proposedItem: (id) item
	     proposedChildIndex: (int) index
{
  if ( ![item respondsToSelector: @selector(childCount)] ||
       index < 0 || index >= [(FolderNode*)item childCount] )
    {
      return NSDragOperationNone;
    }
   
  // Let's get the right item..
  item = [item childAtIndex: index];

  if ([info draggingSourceOperationMask] & NSDragOperationGeneric)
    {
      [theOutlineView setDropItem: item
		      dropChildIndex: NSOutlineViewDropOnItemIndex];
      return NSDragOperationGeneric;
    }
  else if ([info draggingSourceOperationMask] & NSDragOperationCopy)
    {
      [theOutlineView setDropItem: item
		      dropChildIndex: NSOutlineViewDropOnItemIndex];
      return NSDragOperationCopy;
    }
  else
    {
      return NSDragOperationNone;
    }
}


//
// NSOutlineViewDataSource Drag and drop
//
- (BOOL) outlineView: (NSOutlineView*) outlineView
	  acceptDrop: (id <NSDraggingInfo>) info
		item: (id) item
	  childIndex: (int) index
{
  Folder *aSourceFolder, *aDestinationFolder;
  Store *aSourceStore, *aDestinationStore;
  MailWindowController *aMailWindowController;

  FolderNode *aFolderNode;
  NSString *aFolderName;
  NSArray *propertyList;
  
  NSDictionary *aDictionary;
  Message *aMessage;
  int i, count, op;
  
  if (!item || index != NSOutlineViewDropOnItemIndex)
    {
      NSBeep();
      return NO;
    }
  
  aFolderNode = (FolderNode *)item;

  // We get our store and our folder name
  aDestinationStore = [self storeForFolderNode: aFolderNode];

  aFolderName = [Utilities pathOfFolderFromFolderNode: aFolderNode
			   separator: [(id<Store>)aDestinationStore folderSeparator]];
  
  // We get the MailWindowController source
  aMailWindowController = (MailWindowController *)[[info draggingSource] delegate];
  
  if (!aMailWindowController || ![aMailWindowController isKindOfClass: [MailWindowController class]] ||
      !aFolderName || [aFolderName length] == 0)
    {
      NSBeep();
      return NO;
    }
  
  // We verify if we aren't trying to transfer to the current mbox!
  aSourceFolder = [aMailWindowController folder];
  aSourceStore = [aSourceFolder store];
  
  if (aSourceStore == aDestinationStore && [[aSourceFolder name] isEqualToString: aFolderName])
    {
      NSRunInformationalAlertPanel(_(@"Transfer error!"),
				   _(@"You cannot transfer a message inside the same mailbox!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return NO;      
    }


  // We get a reference to our destination folder, w/o parsing it if it's not already open.
  // or w/o selecting it if it's an IMAP store.
  if ([(id<NSObject>)aDestinationStore isKindOfClass: [IMAPStore class]])
    {
      NSLog(@"GETTING IMAP FODLER %@", aFolderName);
      aDestinationFolder = (Folder *)[(IMAPStore *)aDestinationStore folderForName: aFolderName
						   select: NO];
    }
  else
    {
      aDestinationFolder = (Folder *)[(LocalStore *)aDestinationStore folderForName: aFolderName];
      [(LocalFolder *)aDestinationFolder parse];
    }

  [aDestinationFolder setProperty: [NSDate date]  forKey: @"EXPIRE_DATE"];


  // We retrieve property list of messages from paste board
  propertyList = [[info draggingPasteboard] propertyListForType: MessagePboardType];
  
  if (!propertyList)
    {
      return NO;
    }

  count = [propertyList count];
  
  for (i = 0; i < count; i++)
    {
      aDictionary = (NSDictionary *)[propertyList objectAtIndex: i];
      aMessage = [aSourceFolder->allMessages objectAtIndex: [[aDictionary objectForKey: @"MessageNumber"] intValue]-1];
      op = (([info draggingSourceOperationMask]&NSDragOperationGeneric) == NSDragOperationGeneric ? MOVE_MESSAGES : COPY_MESSAGES);

      [self transferMessages: [NSArray arrayWithObject: aMessage]
	    fromStore: aSourceStore
	    fromFolder: aSourceFolder
	    toStore: aDestinationStore
	    toFolder: aDestinationFolder
	    operation: op];
    }

  return YES;
}


//
//
//
#ifndef MACOSX
- (void)    outlineView: (NSOutlineView *) aOutlineView
 willDisplayOutlineCell: (id) aCell
         forTableColumn: (NSTableColumn *) aTbleColumn
                   item: (id)item
{
  if (![aOutlineView isExpandable: item])
    {
      [aCell setImage: nil];
    }
  else
    {
      if ([aOutlineView isItemExpanded: item])
	{
	  [aCell setImage: [NSImage imageNamed: @"sort_down"]];
	}
      else
	{
	  [aCell setImage: [NSImage imageNamed: @"sort_right"]];
	}
    }
}
#endif


//
//
//
- (void) windowDidLoad
{
  NSMenuItem *aMenuItem;
  
  menu = [[NSMenu alloc] init];
  [menu setAutoenablesItems: NO];
  
  aMenuItem = [[NSMenuItem alloc] initWithTitle: _(@"Delete") action: @selector(delete:)  keyEquivalent: @""];
  [aMenuItem setTarget: self];
  [menu addItem: aMenuItem];
  RELEASE(aMenuItem);
  
  aMenuItem = [[NSMenuItem alloc] initWithTitle: _(@"Create") action: @selector(create:)  keyEquivalent: @""];
  [aMenuItem setTarget: self];
  [menu addItem: aMenuItem];
  RELEASE(aMenuItem);
  
  aMenuItem = [[NSMenuItem alloc] initWithTitle: _(@"Rename") action: @selector(rename:)  keyEquivalent: @""];
  [aMenuItem setTarget: self];
  [menu addItem: aMenuItem];
  RELEASE(aMenuItem);

  [self setMailboxManagerCache: [MailboxManagerCache mailboxManagerCacheFromDisk]];

  // We initialize our array containing all Stores and we load of folders
  _allFolders = [[NSMutableArray alloc] init];
  
  // We initialize our dictionary containing all openend IMAPStores
  allStores = [[NSMutableDictionary alloc] init];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_accountsHaveChanged:)
    name: AccountsHaveChanged
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderCreateCompleted:)
    name: @"PantomimeFolderCreateCompleted"
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderCreateFailed:)
    name: @"PantomimeFolderCreateFailed"
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderDeleteCompleted:)
    name: @"PantomimeFolderDeleteCompleted"
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderDeleteFailed:)
    name: @"PantomimeFolderDeleteFailed"
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderRenameCompleted:)
    name: @"PantomimeFolderRenameCompleted"
    object: nil];
  
  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderRenameFailed:)
    name: @"PantomimeFolderRenameFailed"
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderSubscribeCompleted:)
    name: @"PantomimeFolderSubscribeCompleted"
    object: nil];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(_folderUnsubscribeCompleted:)
    name: @"PantomimeFolderUnsubscribeCompleted"
    object: nil];
}


//
// action methods
//
- (IBAction) open: (id) sender
{
  id item;
  int row, level;
  
  row = [outlineView selectedRow];

  // If no row is selected and we had a last MailWindow on top, we set its folder to nil
  if (row < 0 && [GNUMail lastMailWindowOnTop])
    {
      [[[[GNUMail lastMailWindowOnTop] windowController] folder] close];

      if ([[[GNUMail lastMailWindowOnTop] windowController] isKindOfClass: [MailWindowController class]])
	{
	  [[[GNUMail lastMailWindowOnTop] windowController] setFolder: nil];
	}
      else
	{
	  [[[[GNUMail lastMailWindowOnTop] windowController] mailWindowController] setFolder: nil];
	}
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  //
  // We must verify that:
  //
  // a) we have at least one selected row
  // b) we haven't selected our root, or a store (local or IMAP)
  //
  if ([outlineView numberOfSelectedRows] != 1)
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid mailbox to open!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
  else if (level < 1)
    {
      if (![outlineView isItemExpanded: item])
	{
	  [outlineView expandItem: item];
	}
      return;
    }
 
  
  // We verify if it's a local folder
  if ([[Utilities completePathForFolderNode: item  separator: @"/"] 
	hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]])
    { 
      NSString *aString;
      
      aString = [Utilities pathOfFolderFromFolderNode: (FolderNode *)item
			   separator: @"/"];

      [self _openLocalFolderWithName: aString  sender: sender];
    }
  // It's an IMAP folder...
  else
    {
      NSString *aServerName, *aUsername, *aString;
      IMAPStore *aStore;

      [Utilities storeKeyForFolderNode: item
      		 serverName: &aServerName
      		 username: &aUsername];

      aStore = (IMAPStore *)[self storeForName: aServerName
				  username: aUsername];
      
      aString = [[Utilities pathOfFolderFromFolderNode: (FolderNode *)item  separator: @"/"]
		  stringByReplacingOccurrencesOfString: @"/"  withString: [aStore folderSeparator]];
      
      [self _openIMAPFolderWithName: aString
	    store: aStore
	    sender: sender];
    }
}


//
//
//
- (IBAction) create: (id) sender
{
  NewMailboxPanelController *theController;
  id aStore, item;

  int row, level;
  
  row = [outlineView selectedRow];

  if (row < 0)
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  if ([outlineView numberOfSelectedRows] != 1)
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid root where to create this new mailbox."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
  
  //
  // We create our NewMailboxPanelController object. It'll be automatically deallocated when the 
  // window will be closed.
  //
  theController = [[NewMailboxPanelController alloc] initWithWindowNibName: @"NewMailboxPanel"];

  //
  // We get the right store and we disable our mailbox type popup button 
  // if we are creating an IMAP mailbox.
  //
  aStore = [self storeForFolderNode: item];

  [[theController popUpButton] setEnabled: [(id<NSObject>)aStore isKindOfClass: [LocalStore class]]];
  
  if ([NSApp runModalForWindow: [theController window]] == NSRunStoppedResponse)
    {
      NSString *aString, *pathOfFolder;

      pathOfFolder = [Utilities pathOfFolderFromFolderNode: item
				separator: [aStore folderSeparator]];
      
      NSLog(@"pathOfFolder = |%@|", pathOfFolder);
      
      if (!pathOfFolder || [pathOfFolder length] == 0)
	{
	  aString = [[[theController mailboxNameField] stringValue] stringByTrimmingWhiteSpaces];
	}
      else
	{
	  aString = [NSString stringWithFormat: @"%@%@%@",
			      pathOfFolder,
			      [aStore folderSeparator],
			      [[[theController mailboxNameField] stringValue] stringByTrimmingWhiteSpaces]];
	}
      
      NSLog(@"new pathOfFolder = |%@|", aString);

      // We can now proceed with the creation of our new folder
      [aStore createFolderWithName: aString 
	      type: [[theController popUpButton] indexOfSelectedItem]
	      contents: nil];
    }
 
#ifndef MACOSX
  [[self window] makeKeyAndOrderFront: self];
#endif

  RELEASE(theController);
}


//
//
//
- (IBAction) delete: (id) sender
{
  int choice, row, level;
  id item;
  
  row = [outlineView selectedRow];

  if ( row < 0 )
    {
      NSBeep();
      return;
    }

  item = [outlineView itemAtRow: row];
  level = [outlineView levelForItem: item];
  
  if ([outlineView numberOfSelectedRows] != 1 || level < 1)
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"Please select the mailbox you would like to delete."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
      return;
    }
  
  
  // We show our prompt panel
  choice = NSRunAlertPanel(_(@"Delete..."),
			   _(@"Are you sure you want to delete this mailbox?"),
			   _(@"Delete"),  // default
			   _(@"Cancel"),  // alternate
			   nil);
  
  if (choice == NSAlertDefaultReturn)
    {
      NSString *aDefaultMailbox, *aFolderName, *aString;
      id aStore;
             
      aString = [self _stringValueOfURLNameFromItem: item  store: &aStore];
      
      // We get our folder name, respecting the folder separator
      aFolderName = [Utilities pathOfFolderFromFolderNode: (FolderNode *)item
			       separator: [(id<Store>)aStore folderSeparator]];
      
      if ([self _deletingDefaultMailbox: &aDefaultMailbox  usingURLNameAsString: aString])
	{
	  NSRunAlertPanel(_(@"Error while deleting!"),
  			  _(@"You can't delete your default %@ mailbox. Use the Mailboxes tab in the\nAccount Preferences panel to change it before trying again."),
			  _(@"OK"),   // default
			  NULL,       // alternate
  			  NULL,
			  aDefaultMailbox);
	  return;
	}
      
      if ([aStore folderForNameIsOpen: aFolderName])
	{
	  id aWindow;

	  // Get the associated MailWindow.
	  aWindow = [Utilities windowForFolderName: aFolderName  store: aStore];
	  
	  // We just close the mailbox and leave its MailWindow empty!
	  [[[aWindow windowController] folder] close];
	  [[aWindow windowController] setFolder: nil];
	}
 
      // We now delete the mailbox...
      [aStore deleteFolderWithName: aFolderName];
    }
}


//
//
//
- (IBAction) rename: (id) sender
{
  [outlineView editColumn: 0
	       row: [outlineView selectedRow]
	       withEvent: nil
	       select: YES];
}


//
// access / mutation methods
//
- (NSOutlineView *) outlineView
{
  return outlineView;
}


//
// This method returns a pointer to an open store which has
// theFolderNode as one of its children.
//
- (id) storeForFolderNode: (FolderNode *) theFolderNode
{
  Store *aStore;
  
  if ( [[Utilities completePathForFolderNode: theFolderNode  separator: @"/"]
	 hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    {
      aStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"  username: NSUserName()];
    }
  else
    {
      NSString *aServerName, *aUsername;
      
      [Utilities storeKeyForFolderNode: theFolderNode
		 serverName: &aServerName
		 username: &aUsername];
      
      aStore = [self storeForName: aServerName  username: aUsername];
    }

  return aStore;
}


//
//
//
- (id) storeForName: (NSString *) theName
	   username: (NSString *) theUsername
{
  NSString *aString;
  
  aString = [NSString stringWithFormat: @"%@ @ %@", theUsername, theName];
  
  return [allStores objectForKey: aString];
}


//
//
//
- (id) storeForURLName: (URLName *) theURLName
{
  id aStore;
  
  if ([[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame )
    {
      aStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
		     username: NSUserName()];
    }
  else
    {
      if ([self _initializeIMAPStoreWithAccountName: [Utilities accountNameForServerName: [theURLName host]
								username: [theURLName username]]])
	{
	  aStore = [self storeForName: [theURLName host]
			 username: [theURLName username]];
	}
      else
	{
	  aStore = nil;
	}
    }
  
  return aStore;
}


//
//
//
- (void) setStore: (id) theStore
	     name: (NSString *) theName
	 username: (NSString *) theUsername
{
  NSString *aString;

  aString = [NSString stringWithFormat: @"%@ @ %@", theUsername, theName];

  // We verify if we want to remove an opened store.
  if ( !theStore && theName && theUsername )
    {
      FolderNode *aFolderNode;
      int row;

      // For an IMAP store, we remove all children of our root node
      aFolderNode = [self _storeFolderNodeForName: [Utilities accountNameForServerName: theName  username: theUsername]];

#ifndef MACOSX
      [aFolderNode setChildren: nil];
#endif
      [outlineView collapseItem: aFolderNode];
      
      row = [outlineView rowForItem: aFolderNode];

      if (row >= 0 && row < [outlineView numberOfRows])
	{
	  [outlineView selectRow: row
		       byExtendingSelection: NO];
	}

      [allStores removeObjectForKey: aString];
      return;
    }
  
  // We always first "remove" the object in case we call this method
  // multiple times with the same object.
  RETAIN(theStore);
  [allStores removeObjectForKey: aString];
      
  // We {re}add it to our dictionary.
  [allStores setObject: theStore  forKey: aString];
  RELEASE(theStore);
}


//
//
//
- (MailboxManagerCache *) mailboxManagerCache
{
  return mailboxManagerCache;
}


//
//
//
- (void) setMailboxManagerCache: (MailboxManagerCache *) theMailboxManagerCache
{
  ASSIGN(mailboxManagerCache, theMailboxManagerCache);
}


//
//
//
- (void) panic: (NSData *) theData
	folder: (NSString *) theFolder
{
  LocalStore *aLocalStore;
  LocalFolder *aFolder;

  
  NSRunAlertPanel(_(@"Error!"),
		  _(@"An error occured while trying to open the %@ mailbox. This mailbox was probably\ndeleted and a filter is trying to save mails in it. Check your filters and the special mailboxes for all accounts.\nThe message has been saved in the \"Panic\" local mailbox."),
		  _(@"OK"),
		  NULL,
		  NULL,
		  theFolder);


  aLocalStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"  username: NSUserName()]; 

  if (![[NSFileManager defaultManager] fileExistsAtPath: [[aLocalStore path] stringByAppendingPathComponent: @"Panic"]])
    {
      [aLocalStore createFolderWithName: @"Panic"  type: PantomimeFormatMbox  contents: nil];
    }

  aFolder = [aLocalStore folderForName: @"Panic"];
  [aFolder parse];
  [aFolder appendMessageFromRawSource: theData  flags: nil];
}


//
// This method appends a message to the folder specified in theURLName.
//
- (BOOL) addMessage: (NSData *) theMessage
	   toFolder: (URLName *) theURLName
{
  NSString *aFolderName;
  Store *aStore;
  Folder *aFolder;
 
  BOOL storeWasOpen;
   
#warning remove that, we do not need to close the store since they must be always open
  // We verify if out store was open. If it wasn't, we will close the store at the end of this method.
  storeWasOpen = [self storeWithURLNameIsOpen: theURLName];

  aStore = [self storeForURLName: theURLName];
  aFolderName = [theURLName foldername];
  
  if (!aStore)
    {
      [self panic: theMessage  folder: aFolderName];
      return NO;
    }

  if ([(id<NSObject>)aStore isKindOfClass: [IMAPStore class]])
    {
      aFolder = (Folder *)[(IMAPStore *)aStore folderForName: aFolderName  select: NO];
    }
  else
    {
      aFolder = (Folder *)[(LocalStore *)aStore folderForName: aFolderName];
      [(LocalFolder *)aFolder parse];
    }

  if (!aFolder)
    {
      [self panic: theMessage  folder: aFolderName];
      return NO; 
    }

  [aFolder setProperty: [NSDate date]  forKey: @"EXPIRE_DATE"];

  if (![self transferMessage: theMessage
	     flags: nil
	     folder: aFolder])
    {
      NSRunAlertPanel(_(@"Error!"),
		      _(@"An error occured while adding the message to the %@ folder."),
		      _(@"OK"),
		      NULL,
		      NULL,
		      aFolderName);
    }

  //
  // We verify if we must close the store. That could only happen with IMAP stores since the
  // local store is always open.
  //
  if (!storeWasOpen)
    {
      [(IMAPStore *)aStore close];
      [self setStore: nil
	    name: [theURLName host]
	    username: [theURLName username]];
    }

  return YES;
}


//
// 
//
- (Message *) messageFromDraftsFolder
{
  id aMailWindowController;
  Message *aMessage;
  
  aMailWindowController = [[GNUMail lastMailWindowOnTop] delegate];

  // We first verify if current folder is a Drafts folder.
  if ( aMailWindowController && [aMailWindowController isKindOfClass: [MailWindowController class]] )
    {
      if ( ![Utilities stringValueOfURLName: [Utilities stringValueOfURLNameFromFolder: 
							  [aMailWindowController folder]]
		       isEqualTo: @"DRAFTSFOLDERNAME"] )
	{
	  return nil;
	}
    }
   
  if ( [[aMailWindowController folder] count] > 0 &&
       [aMailWindowController selectedMessage])
    {
      aMessage = [aMailWindowController selectedMessage];
    }
  else
    {
      aMessage = nil;
    }

  return aMessage;
}


//
//
//
- (NSDictionary *) allStores
{
  return [NSDictionary dictionaryWithDictionary: allStores];
}

 
//
// This method is used under OS X since we must "swap" the current
// outline view in the Mailbox Manager to match the one currently
// on top in the NSDrawer. This is needed so -open: and other methods
// can work properly with the "current" outline view.
//
#ifdef MACOSX
- (void) setCurrentOutlineView: (id) theOutlineView
{
  outlineView = theOutlineView;
}
#endif


//
//
//
- (void) updateFolderInformation: (NSDictionary *) theInformation
{
  FolderInformation *aFolderInformation;
  NSString *aFolderName;


  aFolderInformation = [theInformation objectForKey: @"FOLDER_INFORMATION"];
  aFolderName = [[theInformation objectForKey: @"FOLDER_NAME"] stringByReplacingOccurrencesOfString: [theInformation objectForKey: @"FOLDER_SEPARATOR"]
							       withString: @"/"];

  [mailboxManagerCache setAllValuesForStoreName: [theInformation objectForKey: @"STORE_NAME"]
		       folderName: aFolderName 
		       username: [theInformation objectForKey: @"USERNAME"]
		       nbOfMessages: [aFolderInformation nbOfMessages]
		       nbOfUnreadMessages: [aFolderInformation nbOfUnreadMessages]];
  
  //
  // We now get the right outline view item to refresh. This is considerably faster
  // than calling -setNeedsDisplay: on the entire outline view (which will result
  // in a plethora of method calls to get the "right values").
  //
  [self updateOutlineViewForFolder: aFolderName
	store: [theInformation objectForKey: @"STORE_NAME"]
	username: [theInformation objectForKey: @"USERNAME"]];
}



//
// This method is used to refresh ONLY the item associated with the Store/Folder.
//
- (void) updateOutlineViewForFolder: (NSString *) theFolder
			      store: (NSString *) theStore
			   username: (NSString *) theUsername
{
  FolderNode *aFolderNode, *aRootNode;
  int row;
  
  if ([theStore isEqualToString: @"GNUMAIL_LOCAL_STORE"])
    {
      aRootNode = localNodes;
    }
  else
    {
      aRootNode = [self _storeFolderNodeForName: [Utilities accountNameForServerName: theStore  username: theUsername]];
    }

  aFolderNode = [Utilities folderNodeForPath: theFolder
			   using: aRootNode
			   separator: @"/"];
  
  row = [outlineView rowForItem: aFolderNode];
  
  if (row >= 0 && row < [outlineView numberOfRows])
    {
      [outlineView setNeedsDisplayInRect: [outlineView rectOfRow: row]];
    }
}


//
// class methods
//
+ (id) singleInstance
{
  if (!singleInstance)
    {
#ifdef MACOSX
      singleInstance = [[MailboxManagerController alloc] init];
#else
      singleInstance = [[MailboxManagerController alloc] initWithWindowNibName: @"MailboxManager"];
#endif
    }
  
  return singleInstance;
}


//
// Other methods
//
- (void) openFolderWithURLName: (URLName *) theURLName
			sender: (id) theSender
{
  if ([[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame)
    {
      [self _openLocalFolderWithName: [theURLName foldername]
	    sender: theSender];
    }
  else if ([[theURLName protocol] caseInsensitiveCompare: @"IMAP"] == NSOrderedSame)
    {
      if ([self _initializeIMAPStoreWithAccountName: [Utilities accountNameForServerName: [theURLName host]
								username: [theURLName username]]])
	{
	  [self _openIMAPFolderWithName: [theURLName foldername]
		store: (IMAPStore *)[self storeForName: [theURLName host]  username: [theURLName username]]
		sender: theSender];
	}
    }
}


//
//
//
- (void) reloadAllFolders
{
  DESTROY(localNodes);

  // We remove all our elements
  [_allFolders removeAllObjects];

  // We add our local folder, if we need to
  localNodes = [Utilities folderNodesFromFolders: [[self storeForName: @"GNUMAIL_LOCAL_STORE"
							 username: NSUserName()] folderEnumerator]
			  separator: @"/"];

  [localNodes setName: _(@"Local")];
  [localNodes setParent: nil];

  if ( [localNodes childCount] > 0 )
    {
      [_allFolders addObject: localNodes];
    }

  RETAIN(localNodes);

  // We verify if the ACCOUNTS preferences have been defined.
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] )
    {
      NSMutableDictionary *allAccounts;
      NSDictionary *allValues;
      NSEnumerator *allKeys;
      NSString *aKey;

      allAccounts = [[NSMutableDictionary alloc] init];
      [allAccounts addEntriesFromDictionary: [Utilities allEnabledAccounts]];
      
      allKeys = [[[allAccounts allKeys] sortedArrayUsingSelector: @selector(compare:)] objectEnumerator];
      
      // We build a correct subset of all our IMAP servers defined in all accounts
      while ( (aKey = [allKeys nextObject]) )
	{
	  allValues = [[allAccounts objectForKey: aKey] objectForKey: @"RECEIVE"];  
	  
	  // We add it only if it's an IMAP server AND if we receive mails either
	  // manually or automatically
	  if ( [[allValues objectForKey: @"SERVERTYPE"] intValue] == IMAP &&
	       [[allValues objectForKey: @"RETRIEVEMETHOD"] intValue] != NEVER )
	    {
	      NSString *aServerName, *aUsername;
	      FolderNode *aFolderNode;
	      
	      aServerName = [allValues objectForKey: @"SERVERNAME"];
	      aUsername = [allValues objectForKey: @"USERNAME"];
	      
	      aFolderNode = [FolderNode folderNodeWithName: [NSString stringWithFormat: @"%@", aKey]
					parent: nil];
	      
	      [_allFolders addObject: aFolderNode];
	      
	      // If our IMAP Store has been previously initialized, we re-initialize it in order to get
	      // the most recent values for the subscribed folders
	      if ( [self storeForName: aServerName
			 username: aUsername] )
		{
		  [self _reloadFoldersForIMAPStoreWithName: aServerName
			username: aUsername
			account: aKey];
		}
	    }
	}
      
      RELEASE(allAccounts);
    }
  
  // We inform our outline view to reload its data.
  [outlineView reloadData];

  // We alwas expand our root item
  [outlineView expandItem: _allFolders];
  
  // We now select and expand the 'Local' folder if there's no IMAP folders defined
  if ([_allFolders count] == 1)
    {
      [outlineView expandItem: localNodes];
      [outlineView selectRow: [outlineView rowForItem: localNodes]
  		   byExtendingSelection: NO];
   }
}


//
//
//
- (BOOL) transferMessage: (NSData *) theMessage
		   flags: (Flags *) theFlags
		  folder: (Folder *) theFolder
{
  Flags *flags;
  id aWindow;

  if (!theMessage || [theMessage length] == 0)
    {
      return NO;
    }
  
  //
  // We transfer the message. If we are transferring to a Sent folder, mark it as read.
  //
  flags = theFlags;

  if ([Utilities stringValueOfURLName: [Utilities stringValueOfURLNameFromFolder: theFolder]  
		 isEqualTo: @"SENTFOLDERNAME"])
    {
      flags = [[Flags alloc] initWithFlags: PantomimeSeen];
      AUTORELEASE(flags);
    }

  aWindow = [Utilities windowForFolderName: [theFolder name]
		       store: [theFolder store]];

  [theFolder appendMessageFromRawSource: theMessage  flags: flags];

#warning move this to folderAppendCompleted?
  if (!aWindow)
    {
      //
      // We update the Mailbox Manager cache, IIF the folder wasn't open initially.
      // If it was open, we had its window and we are gonna refresh its cache in
      // -updateStatusLabel.
      //
      NSString *aFolderName, *aUsername, *aStoreName;
      int nbOfMessages, nbOfUnreadMessages;
      
      aFolderName = [theFolder name];

      if ([theFolder isKindOfClass: [LocalFolder class]])
	{
	  nbOfMessages = [theFolder count];
	  nbOfUnreadMessages = [theFolder numberOfUnreadMessages];
	  aStoreName = @"GNUMAIL_LOCAL_STORE";
	  aUsername = NSUserName();
	}
      else
	{      
	  //
	  // If we are transferring to an IMAP folder, we must obtain the
	  // current values of the target folder since we haven't SELECT'ed it
	  // (so [theFolder count] and [theFolder numberOfUnreadMessages] return 0.
	  //
	  aStoreName = [(IMAPStore *)[theFolder store] name];
	  aUsername = [(IMAPStore *)[theFolder store] username];
	  [mailboxManagerCache allValuesForStoreName: aStoreName
			       folderName: [aFolderName stringByReplacingOccurrencesOfString: [(IMAPStore *)[theFolder store] folderSeparator]
							withString: @"/"]
			       username: aUsername
			       nbOfMessages: &nbOfMessages
			       nbOfUnreadMessages: &nbOfUnreadMessages];
	  
	  if (flags && ![flags contain: PantomimeSeen]) nbOfUnreadMessages += 1;
	  nbOfMessages += 1;
	}
      
      [mailboxManagerCache setAllValuesForStoreName: aStoreName
			   folderName: [aFolderName stringByReplacingOccurrencesOfString: [(IMAPStore *)[theFolder store] folderSeparator]
						    withString: @"/"]
			   username: aUsername
			   nbOfMessages: nbOfMessages
			   nbOfUnreadMessages: nbOfUnreadMessages];
      
      [self updateOutlineViewForFolder: aFolderName
	    store: aStoreName
	    username: aUsername];
    }
  else
    {
      [[aWindow delegate] tableViewShouldReloadData];
      [[aWindow delegate] updateStatusLabel];
    }

  return YES;
}


//
// theOperation == COPY / MOVE
// returns the numbers of transferred messages, -1 on error.
//
- (int) transferMessages: (NSArray *) theMessages
	       fromStore: (id) theSourceStore
	      fromFolder: (id) theSourceFolder
		 toStore: (id) theDestinationStore
		toFolder: (id) theDestinationFolder
	       operation: (int) theOperation
{
  int nbOfTransferredMessages;

  if (!theMessages || [theMessages count] == 0 || !theSourceFolder || !theDestinationFolder )
    {
      NSBeep();
      return 0;
    }
  
  nbOfTransferredMessages = 0;
  
  //
  // If we are transferring messages from an IMAPFolder to an IMAPFolder on the SAME
  // IMAPStore, let's use Pantomime's IMAPFolder: -copyMessage: toFolder: method
  // since the operation is gonna be server-side - so MUCH FASTER.
  //
  if ([theSourceStore isKindOfClass: [IMAPStore class]] && theSourceStore == theDestinationStore)
    {
      int nbOfUnreadTransferredMessages;

      nbOfUnreadTransferredMessages = 0;
      
      [theSourceFolder copyMessages: theMessages
		       toFolder: [[(IMAPFolder *)theDestinationFolder name] 
				   stringByReplacingOccurrencesOfString: @"/"  withString: [theDestinationStore folderSeparator]]];
      nbOfTransferredMessages = [theMessages count];
      
      // If we are moving the messages, mark them as deleted.
      if (theOperation == MOVE_MESSAGES)
	{
	  Message *aMessage;
	  Flags *theFlags;
	  int i, count;
	  
	  count = [theMessages count];
	  for (i = 0; i < count; i++)
	    {
	      aMessage = [theMessages objectAtIndex: i];
	      theFlags = [[aMessage flags] copy];
	      
	      // If we are transferring an unread message, we increment our counter
	      if (![theFlags contain: PantomimeSeen]) nbOfUnreadTransferredMessages += 1;
	      
	      [theFlags add: PantomimeDeleted];
	      [aMessage setFlags: theFlags];
	      RELEASE(theFlags);
	    }
	}
      
      //
      // We now update our MailboxManagerController cache. Since the operation
      // was server-side, we must obtain the current values in the cache and
      // increment them according to what has been transferred.
      //
      if (nbOfTransferredMessages > 0)
	{
	  int nbOfMessages, nbOfUnreadMessages;

	  [mailboxManagerCache allValuesForStoreName: [(IMAPStore *)theDestinationStore name]
			       folderName: [[(IMAPFolder *)theDestinationFolder name] 
					     stringByReplacingOccurrencesOfString: [theDestinationStore folderSeparator]
					     withString: @"/"]
			       username: [theDestinationStore username]
			       nbOfMessages: &nbOfMessages
			       nbOfUnreadMessages: &nbOfUnreadMessages];
	  
	  nbOfUnreadMessages += nbOfUnreadTransferredMessages;
	  nbOfMessages += nbOfTransferredMessages;
	  
	  [mailboxManagerCache setAllValuesForStoreName: [(IMAPStore *)theDestinationStore name]
			       folderName: [[(IMAPFolder *)theDestinationFolder name]
					     stringByReplacingOccurrencesOfString: [theSourceStore folderSeparator]
					     withString: @"/"]
			       username: [theDestinationStore username]
			       nbOfMessages: nbOfMessages
			       nbOfUnreadMessages: nbOfUnreadMessages];
	}
    }
  //
  // We are NOT doing an IMAP-to-IMAP (on the same IMAPStore) copy.
  // 
  else
    {
      NSAutoreleasePool *pool;
      Message *aMessage;
      NSData *aData;
      int i;
	  
      for (i = 0; i < [theMessages count]; i++)
	{
	  pool = [[NSAutoreleasePool alloc] init];
	  
	  aMessage = [theMessages objectAtIndex: i];
	  aData = [aMessage rawSource];
	  
	  if (aData)
	    {
	      Flags *theFlags;
	      
	      // We get our flags but we remove the PantomimeDeleted flag from them
	      theFlags = [[aMessage flags] copy];
	      [theFlags remove: PantomimeDeleted];
	      
	      if ([self transferMessage: aData
			flags: AUTORELEASE([theFlags copy])
			folder: theDestinationFolder])
		{
		  // We verify if we flag it as PantomimeDeleted or not (in case of a
		  // copy operation).
		  if ( theOperation == MOVE_MESSAGES )
		    {
		      [theFlags add: PantomimeDeleted];
		      [aMessage setFlags: theFlags];
		    }

		  nbOfTransferredMessages++;
		}
	      RELEASE(theFlags);
	    }
	  else
	    {
	      NSLog(@"RAW SOURCE NOT AVAILABLE RIGHT NOW, SKIP THIS MESSAGE!");
	      
	      // We write the properties so we know we must transfer it once it's loaded
	      [aMessage setProperty: theDestinationStore  forKey: @"DestinationStore"];
	      [aMessage setProperty: theDestinationFolder  forKey: @"DestinationFolder"];
	      [aMessage setProperty: [NSNumber numberWithInt: theOperation]  forKey: @"Operation"];
	    }
	  
	  RELEASE(pool);
	}
    }

  //
  // We refresh our views, if we need to.
  //
    
  // We need to reload our tableView's data if we transferred some messages
  if (nbOfTransferredMessages > 0)
    {
      id aSourceController, aDestinationController;

      aSourceController = [[Utilities windowForFolderName: [(Folder *)theSourceFolder name]  store: theSourceStore] windowController];
      aDestinationController = [[Utilities windowForFolderName: [(Folder *)theDestinationFolder name]  store: theDestinationStore] windowController];

      // We force the reload of our cache in the source folder, no matter what.
      [[aSourceController folder] updateCache];
      [aSourceController tableViewShouldReloadData];
      [aSourceController updateStatusLabel];

      // We also do this for our destination controller if we need to. If we don't,
      // that means there's no associated MailWindow and we simply update the
      // right row in the mailboxes window.
      if (aDestinationController)
	{
	  [[aDestinationController folder] updateCache];
	  [aDestinationController tableViewShouldReloadData];
	  [aDestinationController updateStatusLabel];
	}
      else
	{
	  [self updateOutlineViewForFolder: [(Folder *)theDestinationFolder name]
		store: ([theDestinationStore isKindOfClass: [IMAPStore class]] ? [(IMAPStore *)theDestinationStore name] : @"GNUMAIL_LOCAL_STORE")
		username: ([theDestinationStore isKindOfClass: [IMAPStore class]] ? [theDestinationStore username] : NSUserName())];
	}
    }

  return nbOfTransferredMessages;
}


//
// This method is used to verify if a folder with name "theFolderName"
// exists in the specified local store.
//
- (BOOL) folderNameExists: (NSString *) theFolderName
		 forStore: (Store *) theStore
{
  NSEnumerator *anEnumerator;
  NSString *aString;

  anEnumerator = [(id<Store>)theStore folderEnumerator];

  while ((aString = [anEnumerator nextObject]))
    {
      if ([aString isEqualToString: theFolderName])
	{
	  return YES;
	}
    }
  
  return NO;
}


//
//
//
- (void) reloadFoldersForStore: (id) theStore
		       folders: (NSEnumerator *) theFolders
{
  NSMutableDictionary *allAccounts, *allValues, *theAccount;
  FolderNode *aFolderNode, *nodes;
  NSString *theAccountName;
  NSString *theSeparator;
  NSArray *allFolders;

  aFolderNode = [self _storeFolderNodeForName: [Utilities accountNameForServerName: [(Service *)theStore name]
							  username: [theStore username]]];

  theSeparator = RETAIN([theStore folderSeparator]);
  
  allFolders = [NSArray arrayWithArray: [theFolders allObjects]];

  nodes = [Utilities folderNodesFromFolders: [allFolders objectEnumerator]
		     separator: theSeparator];
  RETAIN(nodes);
  
  [aFolderNode setChildren: [nodes children]];
  
  RELEASE(theSeparator);
  RELEASE(nodes);

#warning optimize by reloading only the item
  [outlineView reloadData];
  [outlineView expandItem: aFolderNode];

  //
  // We finally cache all/subscribed folders in the user's defaults for this account
  //
  theAccountName = [Utilities accountNameForServerName: [(IMAPStore *)theStore name]  username: [theStore username]];
  allAccounts = [[NSMutableDictionary alloc] initWithDictionary: [[NSUserDefaults standardUserDefaults] 
								   objectForKey: @"ACCOUNTS"]];
  theAccount = [[NSMutableDictionary alloc] initWithDictionary: [allAccounts objectForKey: theAccountName]];
  allValues = [[NSMutableDictionary alloc] initWithDictionary: [theAccount objectForKey: @"RECEIVE"]];
  
  //
  // We write back the information
  //
  [allValues setObject: allFolders  forKey: @"SUBSCRIBED_FOLDERS"];
  [theAccount setObject: allValues  forKey: @"RECEIVE"];
  [allAccounts setObject: theAccount  forKey: theAccountName];
  [[NSUserDefaults standardUserDefaults] setObject: allAccounts  forKey: @"ACCOUNTS"];
  [[NSUserDefaults standardUserDefaults] synchronize];
  
  RELEASE(allValues);
  RELEASE(theAccount);
  RELEASE(allAccounts);
}


//
//
//
- (BOOL) storeWithURLNameIsOpen: (URLName *) theURLName
{
  // The local store is always open.
  if ( [[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame )
    {
      return YES;
    }
  
  if ( [self storeForName: [theURLName host]  username: [theURLName username]] )
    {
      return YES;
    }
  
  return NO;
}


//
//
//
- (void) saveMessageInDraftsFolderForController: (EditWindowController *) theEditWindowController
{
  NSString *theAccountName;
  URLName *theURLName;
  
  // We first update the current message content with the current content of the view
  // and we synchronize our popup button.
  [theEditWindowController updateMessageContentFromTextView];
  [[theEditWindowController accountPopUpButton] synchronizeTitleAndSelectedItem];

  // We get our account name
  theAccountName = [Utilities accountNameForItemTitle: [[theEditWindowController accountPopUpButton] titleOfSelectedItem]];

  // We finally get our URLName object.
  theURLName = [[URLName alloc] initWithString: [[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] objectForKey: theAccountName] 
						   objectForKey: @"MAILBOXES"] objectForKey: @"DRAFTSFOLDERNAME"]
				path: [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"]];
  
  [[MailboxManagerController singleInstance] addMessage: [[theEditWindowController message] dataValue]
					     toFolder: theURLName];

  // If this message is already in the Drafts folder, set the "deleted" flag 
  // of the original message.
  if ([theEditWindowController mode] == GNUMailRestoreFromDrafts)
    {
      Flags *theFlags;
      
      theFlags = [[[theEditWindowController message] flags] copy];
      [theFlags add: PantomimeDeleted];
      [[theEditWindowController message] setFlags: theFlags];
      RELEASE(theFlags);

      // We post our notifications
      [[NSNotificationCenter defaultCenter] postNotificationName: SelectionOfMessageHasChanged
					    object: nil
					    userInfo: nil];
      
      [[NSNotificationCenter defaultCenter] postNotificationName: ReloadMessageList
					    object: nil
					    userInfo: nil];
  }
  
  // We mark the window's document as non-edited
  [[theEditWindowController window] setDocumentEdited: NO];

  RELEASE(theURLName);
}

@end


//
// Private methods
//
@implementation MailboxManagerController (Private)

- (void) _accountsHaveChanged: (id) sender
{
  [self reloadAllFolders];
}


//
//
//
- (BOOL) _deletingDefaultMailbox: (NSString **) theMailboxName
	    usingURLNameAsString: (NSString *) theURLNameAsString
{
  if ( [Utilities stringValueOfURLName: theURLNameAsString  isEqualTo: @"INBOXFOLDERNAME"] )
    {
      *theMailboxName = _(@"Inbox");
      return YES;
    }
  else if ( [Utilities stringValueOfURLName: theURLNameAsString  isEqualTo: @"SENTFOLDERNAME"] )
    {
      *theMailboxName = _(@"Sent");
      return YES;
    }
  else if ( [Utilities stringValueOfURLName: theURLNameAsString  isEqualTo: @"DRAFTSFOLDERNAME"] )
    {
      *theMailboxName = _(@"Drafts");
      return YES;
    }
  else if ( [Utilities stringValueOfURLName: theURLNameAsString  isEqualTo: @"TRASHFOLDERNAME"] )
    {
      *theMailboxName = _(@"Trash");
      return YES;
    }
  
  return NO;
}


//
//
//
- (void) _folderCreateCompleted: (NSNotification *) theNotification
{
  NSString *aString;
  id o;

  // We get the account in order to verify if we must subscribe to the IMAP mailbox
  o = [theNotification object];

  if ([o isKindOfClass: [IMAPStore class]])
    {
      NSLog(@"IMAPSTORE!");

      aString = [Utilities accountNameForServerName: [(IMAPStore *)o name]  username: [o username]];
      
      if ([[[[[Utilities allEnabledAccounts] objectForKey: aString] objectForKey: @"RECEIVE"]
	     objectForKey: @"SHOW_WHICH_MAILBOXES"] intValue] == IMAP_SHOW_SUBSCRIBED_ONLY)
	{
	  NSLog(@"SUBSCRIBING TO THE FOLDER");
	  [o subscribeToFolderWithName: [[theNotification userInfo] objectForKey: @"Name"]];
	  return;
	}
    }
  
  [self _folderSubscribeCompleted: theNotification];
}


//
//
//
- (void) _folderCreateFailed: (NSNotification *) theNotification
{
  NSRunInformationalAlertPanel(_(@"Mailbox error!"),
			       _(@"An error occured while creating the %@ mailbox. This mailbox probably already exists\nor you don't have permission to create it."),
			       _(@"OK"),
			       NULL, 
			       NULL,
			       [[theNotification userInfo] objectForKey: @"Name"]);
}


//
//
//
- (void) _folderDeleteCompleted: (NSNotification *) theNotification
{
  id aStore, item;

  NSLog(@"In _folderDeleteCompleted!");

  aStore = [theNotification object];

#warning FIXME get the right item in case the selection has changed
  item = [outlineView itemAtRow: [outlineView selectedRow]];  

  // Delete cache files, ONLY if this is an IMAP folder!
  if ([aStore isKindOfClass: [IMAPStore class]] )
    {
      NSString *aKey, *cacheFilePath;
      FolderNode *node;
      int i;
      
      aKey = [NSString stringWithFormat: @"%@ @ %@", 
		       [(IMAPStore *)aStore username], 
		       [(IMAPStore *)aStore name]];
      
      cacheFilePath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@", 
				GNUMailUserLibraryPath(),
				[Utilities flattenPathFromString: aKey
					   separator: @"/"],
				[Utilities flattenPathFromString: 
					     [Utilities pathOfFolderFromFolderNode: (FolderNode *)item
							separator: [(id<Store>)aStore folderSeparator]]
					   separator: @"/"]];
      
      // We remove the file
      NS_DURING
	[[NSFileManager defaultManager] removeFileAtPath: cacheFilePath
					handler: nil];
      NS_HANDLER
	// Under GNUstep, if we pass something that can't be converted to a cString
	// to -removeFileAtPath, it throws an exception.
	NSDebugLog(@"Exception occured while removing the cache file.");
      NS_ENDHANDLER;
      
      // We remove the cache file of the children of this folder, if any.
      for (i = 0; i < [(FolderNode *)item childCount]; i++)
	{
	  node = [(FolderNode *)item childAtIndex: i];
	  cacheFilePath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@", 
				    GNUMailUserLibraryPath(),
				    [Utilities flattenPathFromString: aKey
					       separator: @"/"],
				    [Utilities flattenPathFromString: 
						 [Utilities pathOfFolderFromFolderNode: node
							    separator: [(id<Store>)aStore folderSeparator]]
					       separator: @"/"]];
	  
	  NS_DURING
	    [[NSFileManager defaultManager] removeFileAtPath: cacheFilePath
					    handler: nil];
	  NS_HANDLER
	    // Under GNUstep, if we pass something that can't be converted to a cString
	    // to -removeFileAtPath, it throws an exception.
	    NSDebugLog(@"Exception occured while removing the cache file.");
	  NS_ENDHANDLER;
	}

      // We get the account in order to verify if we must subscribe to the IMAP mailbox
      aKey = [Utilities accountNameForServerName: [(IMAPStore *)aStore name]  username: [aStore username]];
      
      if ([[[[[Utilities allEnabledAccounts] objectForKey: aKey] objectForKey: @"RECEIVE"]
	     objectForKey: @"SHOW_WHICH_MAILBOXES"] intValue] == IMAP_SHOW_SUBSCRIBED_ONLY)
	{
	  [aStore unsubscribeToFolderWithName: [[theNotification userInfo] objectForKey: @"Name"]];
	  return;
	}
    }
  
  [self _reloadFoldersAndExpandParentsFromNode: [item parent]
	selectNodeWithPath: [Utilities completePathForFolderNode: [item parent]
				       separator: @"/"]];
}


//
//
//
- (void) _folderDeleteFailed: (NSNotification *) theNotification
{
  NSRunInformationalAlertPanel(_(@"Mailbox error!"),
			       _(@"An error occured while deleting the %@ mailbox. This mailbox is probably already deleted\nor the server does not support deleting open mailboxes."),
			       _(@"OK"),
			       NULL, 
			       NULL,
			       [[theNotification userInfo] objectForKey: @"Name"]);
}


//
//
//
- (void) _folderRenameCompleted: (NSNotification *) theNotification
{
  NSString *aSourceURL, *aDestinationURL, *aString, *aName, *aNewName;
  Store *aStore;
  id aWindow;
  
  aStore = [theNotification object];
  aName = [[theNotification userInfo] objectForKey: @"Name"];
  aNewName = [[theNotification userInfo] objectForKey: @"NewName"];

  // We build our right URLs
  if ( [(id<NSObject>)aStore isKindOfClass: [LocalStore class]] )
    {
      aSourceURL = [NSString stringWithFormat: @"local://%@/%@",
			     [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"],
			     aName];
      aDestinationURL = [NSString stringWithFormat: @"local://%@/%@",
				  [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"],
				  aNewName];
    }
  else
    {
      aSourceURL = [NSString stringWithFormat: @"imap://%@@%@/%@", 
			     [(IMAPStore *)aStore username], 
			     [(IMAPStore *)aStore name],
			     aName];
      aDestinationURL = [NSString stringWithFormat: @"imap://%@@%@/%@", 
				  [(IMAPStore *)aStore username], 
				  [(IMAPStore *)aStore name], 
				  aNewName];
    }
  
  // We update our filters.
  [self _updateFiltersFromOldPath: aSourceURL  toPath: aDestinationURL];
  
  // We update our "MAILBOXES" for all accounts
  [self _updateMailboxesFromOldPath: aSourceURL  toPath: aDestinationURL];
  
  
  //
  // Then, we must verify if we must rename our IMAP cache file.
  //
  if ( [(id<NSObject>)aStore isKindOfClass: [IMAPStore class]] )
    {
      NSString *aKey, *aSourcePath, *aDestinationPath;
      
      // FIXME - buggy? IMAPStore must be updated to update the pathToCache ivar in IMAPCacheManager
      // for an open IMAPFolder.
      aKey = [NSString stringWithFormat: @"%@ @ %@", 
		       [(IMAPStore *)aStore username], 
		       [(IMAPStore *)aStore name]];
      
      aSourcePath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@",
			      GNUMailUserLibraryPath(),
			      [Utilities flattenPathFromString: aKey
					 separator: @"/"],
			      [Utilities flattenPathFromString: aName
					 separator: [(id<Store>)aStore folderSeparator]] ];
      
      aDestinationPath = [NSString stringWithFormat: @"%@/IMAPCache_%@_%@", 
				   GNUMailUserLibraryPath(),
				   [Utilities flattenPathFromString: aKey
					      separator: @"/"],
				   [Utilities flattenPathFromString: aNewName
					      separator: [(id<Store>)aStore folderSeparator]] ];
      
      [[NSFileManager defaultManager] movePath: aSourcePath
				      toPath: aDestinationPath
				      handler: nil];
      
    }
  
  // Success! Let's refresh our MM. The _reloadFoldersAndExpandParentsFromNode::
  // method expects to have the store name before the node's path.
  if ( [(id<NSObject>)aStore isKindOfClass: [LocalStore class]] )
    {
      aString = [NSString stringWithFormat: @"/%@/%@", _(@"Local"), aNewName];
    }
  else
    {
      aString = [NSString stringWithFormat: @"/%@/%@",
			  [Utilities accountNameForServerName: [(IMAPStore *)aStore name]  username: [(IMAPStore *)aStore username]],
			  aNewName];
    }
  
  [self _reloadFoldersAndExpandParentsFromNode: [[outlineView itemAtRow: [outlineView selectedRow]] parent]
	selectNodeWithPath: aString];
  
  // We also refresh our window's title
  aWindow = [Utilities windowForFolderName: aNewName  store: aStore];

  if (aWindow)
    {
      [[aWindow windowController] updateWindowTitle];
    }
}


//
//
//
- (void) _folderRenameFailed: (NSNotification *) theNotification
{
  NSRunInformationalAlertPanel(_(@"Mailbox error!"),
			       _(@"An error occured while renaming the %@ mailbox to %@. This mailbox probably already exists\nor you don't have permission to rename it."),
			       _(@"OK"),
			       NULL, 
			       NULL,
			       [[theNotification userInfo] objectForKey: @"Name"],
			       [[theNotification userInfo] objectForKey: @"NewName"]);
}


//
//
//
- (void) _folderSubscribeCompleted: (NSNotification *) theNotification
{
  NSString *aString;
  id item;
  int row;

#warning FIXME get the right item in case the selection has changed
  row = [outlineView selectedRow];

  if (row < 0) return;
  item = [outlineView itemAtRow: row];
  
  aString = [NSString stringWithFormat: @"%@/%@", [Utilities completePathForFolderNode: item  separator: @"/"],
		      [[theNotification userInfo] objectForKey: @"Name"]];
  
  [self _reloadFoldersAndExpandParentsFromNode: item
	selectNodeWithPath: aString];
}


//
//
//
- (void) _folderUnsubscribeCompleted: (NSNotification *) theNotification
{
  NSString *aString;
  id item;
  int row;

#warning FIXME get the right item in case the selection has changed
  row = [outlineView selectedRow];

  if (row < 0) return;
  item = [outlineView itemAtRow: row];
  
  aString = [NSString stringWithFormat: @"%@/%@", [Utilities completePathForFolderNode: item  separator: @"/"],
		      [[theNotification userInfo] objectForKey: @"Name"]];
  
  [self _reloadFoldersAndExpandParentsFromNode: item
	selectNodeWithPath: aString];
}


//
//
//
- (BOOL) _initializeIMAPStoreWithAccountName: (NSString *) theAccountName
{
  NSNumber *portValue;  
  NSString *aServerName, *aUsername;
  NSDictionary *allValues;
  IMAPStore *aStore;
  
  // We begin by searching in our ACCOUNTS values for the right account.
  // Now, let's get all the receive values
  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
		 objectForKey: theAccountName] objectForKey: @"RECEIVE"];
  
  portValue =  [allValues objectForKey: @"PORT"];
	  
  // We use the default IMAP port if it's not defined.
  if (!portValue)
    {
      portValue = [NSNumber numberWithInt: 143];
    }
  
  // We get our username
  aUsername = [allValues objectForKey: @"USERNAME"];

  // We get our servername
  aServerName = [allValues objectForKey: @"SERVERNAME"];
  
  // We first verify if we haven't already cached our store. If so,
  // we simply return since the Store has already been initialized.
  if ([self storeForName: aServerName  username: aUsername])
    {
      return YES;
    }

  aStore = [[IMAPStore alloc] initWithName: aServerName
			      port: [portValue intValue]];
  [aStore addRunLoopMode: NSEventTrackingRunLoopMode];
  [aStore addRunLoopMode: NSModalPanelRunLoopMode];
  [aStore setUsername: [allValues objectForKey: @"USERNAME"]]; 
  
  [aStore setDelegate: [TaskManager singleInstance]];
  
  ADD_CONSOLE_MESSAGE(_(@"Connecting to IMAP server %@"), aServerName);
  [aStore connectInBackgroundAndNotify];
  
  [self setStore: aStore
	name: aServerName
	username: aUsername];
  
#ifndef MACOSX
  if ([[self window] isVisible])
    {
      [[self window] makeKeyAndOrderFront: self];
    }
#endif

  return YES;
}


//
//
//
- (void) _nbOfMessages: (int *) theNbOfMessages
    nbOfUnreadMessages: (int *) theNbOfUnreadMessages
               forItem: (id) theItem
{
  NSString *aString, *aStoreName, *aFolderName, *aUsername;
  
  aString = [Utilities completePathForFolderNode: theItem
		       separator: @"/"];
  
  if ( [aString hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]] )
    {
      aStoreName = @"GNUMAIL_LOCAL_STORE";
      aFolderName = [Utilities pathOfFolderFromFolderNode: (FolderNode *)theItem
			       separator: @"/"];
      aUsername = NSUserName();
    }
  else
    {
      [Utilities storeKeyForFolderNode: theItem
		 serverName: &aStoreName
		 username: &aUsername];
      
      aFolderName = [Utilities pathOfFolderFromFolderNode: (FolderNode *)theItem
			       separator: @"/"];
    }
  
  [mailboxManagerCache allValuesForStoreName: aStoreName
		       folderName: aFolderName
		       username: aUsername
		       nbOfMessages: theNbOfMessages
		       nbOfUnreadMessages: theNbOfUnreadMessages];
}


//
//
//
- (void) _openLocalFolderWithName: (NSString *) theFolderName
			   sender: (id) theSender
{
  MailWindowController *aMailWindowController;
  LocalStore *localStore;
  id aFolder;
  
  BOOL reusingLastMailWindowOnTop, aMask;
  
  // We get out local store and our folder.
  localStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"  username: NSUserName()];
  aFolder = nil;
  
  // We first verify if the folder is still valid. For example, it could have been
  // deleted (the file) manually while GNUMail.app was running.
  if (![[NSFileManager defaultManager] fileExistsAtPath: [[localStore path] stringByAppendingPathComponent: theFolderName]])
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"The local mailbox %@ does not exist!"),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   theFolderName);
      return;
    }

  // We now verify if it's not a directory (a folder holding folders)
  if (([localStore folderTypeForFolderName: theFolderName] & PantomimeHoldsFolders) == PantomimeHoldsFolders)
    {
      FolderNode *item;
      int i;

      item = [outlineView itemAtRow: [outlineView selectedRow]];
      
      aFolder = AUTORELEASE([[VirtualFolder alloc] initWithName: theFolderName]);
      
      // We add all direct sub-mailboxes
      for (i = 0; i < [item childCount]; i++)
	{
	  id o;

	  o = [localStore folderForName: [Utilities pathOfFolderFromFolderNode: [item childAtIndex: i]  separator: @"/"]];
	  [o parse];
	  [aFolder addFolder: o];
	}
    }

  // If the folder is already open, we "focus" that window
  if ([localStore folderForNameIsOpen: theFolderName])
    {
      NSWindow *aWindow;
      
      aWindow = (NSWindow *)[Utilities windowForFolderName: theFolderName 
				       store: (Store *)localStore];
      if (aWindow)
	{
	  [aWindow orderFrontRegardless];
	  return;
	}
    }
  
  // We must open (or get an open folder) the folder.
  if (!aFolder)
    {
      aFolder = [localStore folderForName: theFolderName];
    }
  
#ifdef MACOSX
  aMask = ([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask) == NSAlternateKeyMask;
#else
  aMask = ([[NSApp currentEvent] modifierFlags] & NSControlKeyMask) == NSControlKeyMask;
#endif

  // If we reuse our window controller...
  if ([theSender isKindOfClass: [NSMenuItem class]] || 
      [GNUMail lastMailWindowOnTop] == nil || 
       theSender == [NSApp delegate] ||
      aMask)
    {
      aMailWindowController = [[MailWindowController alloc] initWithWindowNibName: @"MailWindow"];
      reusingLastMailWindowOnTop = NO;
    }
  else
    {      
      aMailWindowController = [[GNUMail lastMailWindowOnTop] windowController];
      reusingLastMailWindowOnTop = YES;
      
      // We must NOT assume that we got a MailWindowController
      if ([aMailWindowController isKindOfClass: [MessageViewWindowController class]])
	{
	  aMailWindowController = [(MessageViewWindowController *)aMailWindowController mailWindowController];
	}
      
      // We close the previous folder. No need to handle the IMAP timeout
      // as it's handled in IMAPFolder: -close.
      [[aMailWindowController folder] close];
    }
      
  // We update the show/hide deleted/read
  [self _updateShowOrHideReadAndDeletedForFolder: aFolder];
  
  // We set the new folder
  [aMailWindowController setFolder: aFolder];

  // We don't try to parse a VirtualFolder
  if ([aFolder isKindOfClass: [LocalFolder class]])
    {
      [aFolder parse];
    }

  // We we are reusing our window controller, we must always reload the table view
  if (reusingLastMailWindowOnTop && [GNUMail lastMailWindowOnTop])
    {
      [aMailWindowController tableViewShouldReloadData];
    }

  // And we show the window.. 
  [[aMailWindowController window] orderFrontRegardless];
  [[aMailWindowController window] makeKeyAndOrderFront: nil];

  ADD_CONSOLE_MESSAGE(_(@"Local folder %@ opened."), theFolderName);

  // If the "Local" node was collapsed in our MailboxManager, we now expend it
  if (![outlineView isItemExpanded: [self _storeFolderNodeForName: _(@"Local")]])
    {
      [outlineView expandItem: [self _storeFolderNodeForName: _(@"Local")]];
    }
}


//
//
//
- (void) _openIMAPFolderWithName: (NSString *) theFolderName
			   store: (IMAPStore *) theStore
			  sender: (id) theSender
{
  MailWindowController *aMailWindowController;
  IMAPCacheManager *anIMAPCacheManager;
  IMAPFolder *aFolder;
  NSString *aKey;
  
  BOOL reusingLastMailWindowOnTop, aMask;

#ifdef MACOSX
  aMask = ([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask) == NSAlternateKeyMask;
#else
  aMask = ([[NSApp currentEvent] modifierFlags] & NSControlKeyMask) == NSControlKeyMask;
#endif

  // Using IMAP, we currently only allow the user to have one folder open
  // at the time on the same IMAPStore.
  if ([[[theStore openFoldersEnumerator] allObjects] count] > 0)
    {
      id aWindow;

      // We search for one open window (so folder) on the IMAP store
      aWindow = [Utilities windowForFolderName: nil  store: (Store *)theStore];

      // If the folder that the window is 'using' is the same as the one we are trying to open,
      // we simply make this window the key one and order it front. There's no need to try
      // to reopen that folder!
      if ( [[[[aWindow windowController] folder] name] isEqualToString: theFolderName] )
	{
	  [aWindow makeKeyAndOrderFront: self];
	  return;
	}

      // If we are trying to open a new MailWindow using the menu item or if we are reusing
      // a MailWindow but the current on top isn't a window 'using' our IMAP store...
      if ( [theSender isKindOfClass: [NSMenuItem class]] ||
	   aMask ||
	   ([[GNUMail allMailWindows] count] > 1 && [GNUMail lastMailWindowOnTop] != aWindow) )
	{
	  NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				       _(@"A folder (%@) is already open. Please close it first."),
				       _(@"OK"),
				       NULL, 
				       NULL,
				       [(IMAPFolder *)[[theStore openFoldersEnumerator] nextObject] name]);
	  return;
	}
    }
  

  //
  // We verify if we must reuse or not our window controller. The first if () is to
  // verify if we must NOT reuse it.
  //
  if ( [theSender isKindOfClass: [NSMenuItem class]] || 
       [GNUMail lastMailWindowOnTop] == nil || 
       theSender == [NSApp delegate] ||
       aMask )
    {
      aMailWindowController = [[MailWindowController alloc] initWithWindowNibName: @"MailWindow"];
      reusingLastMailWindowOnTop = NO;
    }
  else
    {
      aMailWindowController = [[GNUMail lastMailWindowOnTop] windowController];
      reusingLastMailWindowOnTop = YES;
      
      // We must NOT assume that we got a MailWindowController
      if ( [aMailWindowController isKindOfClass: [MessageViewWindowController class]] )
	{
	  aMailWindowController = [(MessageViewWindowController *)aMailWindowController mailWindowController];
	}
      
      // We close the previous folder. No need to handle the IMAP timeout
      // as it's handled in IMAPFolder: -close.
      [[aMailWindowController folder] close];
    }

  // We send our message to the console saying we are about to open the IMAP folder
  ADD_CONSOLE_MESSAGE(_(@"Opening IMAP folder %@ on %@..."), theFolderName, [theStore name]);
  
  // We get our cache manager for this server / folder
  aKey = [NSString stringWithFormat: @"%@ @ %@", [theStore username], [theStore name]];
  anIMAPCacheManager = [IMAPCacheManager cacheWithPath: [NSString stringWithFormat: @"%@/IMAPCache_%@_%@",
								  GNUMailUserLibraryPath(),
								  [Utilities flattenPathFromString: aKey
									     separator: @"/"],
								  [Utilities flattenPathFromString: theFolderName
									     separator: [theStore folderSeparator]]]];
  //
  // We obtain our folder from the IMAP store.
  //
  aFolder = (IMAPFolder *)[theStore folderForName: theFolderName
				    mode: PantomimeReadWriteMode
				    prefetch: NO];

  // We verify if the folder can be open. It could have been a \NoSelect folder.
  if (!aFolder)
    {
      NSRunInformationalAlertPanel(_(@"Mailbox error!"),
				   _(@"You must select a valid mailbox to open!"),
				   _(@"OK"),
				   NULL,
				   NULL,
				   NULL);
      return;
    }

  //
  // We verify if we haven't got a non-selected folder.
  //
#warning what happens it we are DnD to this folder?
  if (![aFolder selected])
    {
      [aFolder close];

      aFolder = (IMAPFolder *)[theStore folderForName: theFolderName
					mode: PantomimeReadWriteMode
					prefetch: NO];
    }


  // We set the cache manager and we prefetch our messages
  [aFolder setCacheManager: anIMAPCacheManager];
  //[aFolder prefetch];
  

  // We update the show/hide deleted/read
  [self _updateShowOrHideReadAndDeletedForFolder: aFolder];
  
  // We set the folder
  [aMailWindowController setFolder: aFolder];
  
#if 1
#warning do not clear the view but handle when the user click on a message of a closed folder
  // We are reusing our window controller, we must always reload the table view
  if ( reusingLastMailWindowOnTop && [GNUMail lastMailWindowOnTop] )
    {
      [aMailWindowController tableViewShouldReloadData];
    }
#endif

  [[aMailWindowController window] orderFrontRegardless];
  [[aMailWindowController window] makeKeyAndOrderFront: nil];

  ADD_CONSOLE_MESSAGE(_(@"IMAP folder %@ on %@ opened."), theFolderName, [theStore name]);

  // If the "IMAP" node was collapsed in our MailboxManager, we now expand it
  if (![outlineView isItemExpanded: [self _storeFolderNodeForName: 
					    [Utilities accountNameForServerName: [theStore name]
						       username: [theStore username]]]])
    {
      [outlineView expandItem: [self _storeFolderNodeForName: 
				       [Utilities accountNameForServerName: [theStore name]
						  username: [theStore username]]] ];
    }
}


//
//
//
- (void) _reloadFoldersAndExpandParentsFromNode: (FolderNode *) theNode
			     selectNodeWithPath: (NSString *) thePath
{ 
  NSString *aString, *aServerName, *aUsername;
  NSMutableArray *nodesToExpand;
  
  id aParent, aRootNode;
  int i, aRow; 
  
  aString = [Utilities storeKeyForFolderNode: theNode
		       serverName: &aServerName
		       username: &aUsername];

  // We must refresh our outline view by reload its content
  [self reloadAllFolders];
  
  // We first get our root node
  if ([thePath hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]])
    {
      aRootNode = localNodes;
    }
  else
    {
      aRootNode = [self _storeFolderNodeForName: [Utilities accountNameForServerName: aServerName  username: aUsername]];
    }


  // We get our new node in our tree and also our new row index
  aParent = [Utilities folderNodeForPath: [thePath stringByDeletingLastPathComponent]
		       using: aRootNode
		       separator: @"/"];

  nodesToExpand = [[NSMutableArray alloc] init];

  // We expand all our parent, to make the row visible.
  while (aParent)
    {
      [nodesToExpand addObject: aParent];
      aParent = [aParent parent];
    }
  
  // We must expand our nodes starting from the root to the children and not
  // the other way around. Otherwise, the NSOutlineView just borks.
  for (i = ([nodesToExpand count] - 1); i >= 0; i--)
    {
      [outlineView expandItem: [nodesToExpand objectAtIndex: i]];
    }

  RELEASE(nodesToExpand);

  // We now get our new node node (renamed or created). Since it's now shown on the screen,
  // we can now obtain the row for it and select it.
  aParent = [Utilities folderNodeForPath: thePath
		       using: aRootNode
		       separator: @"/"];

  aRow = [outlineView rowForItem: aParent];


  if (aRow >= 0 && aRow < [outlineView numberOfRows])
    {
      [outlineView selectRow: aRow
      		   byExtendingSelection: NO];
      [outlineView scrollRowToVisible: aRow];
    }
}


//
//
//
#warning REMOVE?
- (void) _reloadFoldersForIMAPStoreWithName: (NSString *) theName
				   username: (NSString *) theUsername
				    account: (NSString *) theAccountName
{
  FolderNode *aFolderNode;
  IMAPStore *aStore;
  
  aStore = [self storeForName: theName  username: theUsername];
  aFolderNode = [self _storeFolderNodeForName: theAccountName];
  
  if (aFolderNode)
    {
      NSArray *subscribedFolders;
      NSNumber *aNumber;
      
      NSString *theSeparator;
      FolderNode *nodes;

      aNumber = [[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
		    objectForKey: theAccountName] objectForKey: @"RECEIVE"] objectForKey: @"SHOW_WHICH_MAILBOXES"];

      if (aNumber && [aNumber intValue] == IMAP_SHOW_SUBSCRIBED_ONLY)
	{
	  subscribedFolders = [NSArray arrayWithArray: [[aStore subscribedFolderEnumerator] allObjects]];
	}
      else
	{
	  subscribedFolders = [NSArray arrayWithArray: [[aStore folderEnumerator] allObjects]];
	}
      
      theSeparator = RETAIN([aStore folderSeparator]);

      nodes = [Utilities folderNodesFromFolders: [subscribedFolders objectEnumerator]
			 separator: theSeparator];
      RETAIN(nodes);
      
      [aFolderNode setChildren: [nodes children]];

      RELEASE(theSeparator);
      RELEASE(nodes);
      
      [outlineView reloadItem: aFolderNode];
    }
  else
    {
      NSDebugLog(@"Error getting the folder's root for IMAP folders tree.");
    }
}


//
//
//
- (FolderNode *) _storeFolderNodeForName: (NSString *) theName
{
  int i;

  for (i = 0; i < [_allFolders count]; i++)
    {
      FolderNode *aFolderNode;
      
      aFolderNode = [_allFolders objectAtIndex: i];

      if ( [theName isEqualToString: [aFolderNode name]] )
	{
	  return aFolderNode;
	}
    }

  return nil;
}


//
//
//
- (NSString *) _stringValueOfURLNameFromItem: (id) theItem
				       store: (Store **) theStore
{
  NSMutableString *aMutableString;
  NSString *aString;
  
  aMutableString = [[NSMutableString alloc] init];
  
  // We verify if it's a local folder
  if ([[Utilities completePathForFolderNode: theItem  separator: @"/"] 
	hasPrefix: [NSString stringWithFormat: @"/%@", _(@"Local")]])
    {
      [aMutableString appendFormat: @"local://%@", [[NSUserDefaults standardUserDefaults] 
						     objectForKey: @"LOCALMAILDIR"]];
      *theStore = [self storeForName: @"GNUMAIL_LOCAL_STORE"
			username: NSUserName()];
    }
  else
    {
      NSString *aServerName, *aUsername;
      
      [Utilities storeKeyForFolderNode: theItem
		 serverName: &aServerName 
		 username: &aUsername];
      *theStore = [self storeForName: aServerName
			username: aUsername];
  
      [aMutableString appendFormat: @"imap://%@@%@", aUsername, aServerName];
    }
  
  // We get our folder name, respecting the folder separator
  aString = [Utilities pathOfFolderFromFolderNode: (FolderNode *)theItem
		       separator: [(id<Store>)*theStore folderSeparator]];
  
  [aMutableString appendFormat: @"/%@", aString];

  return AUTORELEASE(aMutableString);
}


//
//
//
- (void) _updateFiltersFromOldPath: (NSString *) theOldPath
			    toPath: (NSString *) thePath
{
  FilterManager *aFilterManager;
  Filter *aFilter;
  int i;
  
  aFilterManager = [FilterManager singleInstance];
  
  for (i = 0; i < [[aFilterManager filters] count]; i++)
    {   
      aFilter = [aFilterManager filterAtIndex: i];
      
      if ([aFilter action] == TRANSFER_TO_FOLDER)
	{
	  if ( [[aFilter actionFolderName] isEqualToString: theOldPath] )
	    {
	      [aFilter setActionFolderName: thePath];
	    }
	}
    }
  
  [aFilterManager synchronize];	
}


//
//
//
- (void) _updateMailboxesFromOldPath: (NSString *) theOldPath
                              toPath: (NSString *) thePath

{
  NSMutableDictionary *allAccounts, *theAccount, *allValues;
  NSEnumerator *theEnumerator;
  NSString *aKey;
  
  allAccounts = [[NSMutableDictionary alloc] initWithDictionary: [[NSUserDefaults standardUserDefaults] 
								   dictionaryForKey: @"ACCOUNTS"]];
  
  theEnumerator = [allAccounts keyEnumerator];
  
  while ( (aKey = [theEnumerator nextObject]) )
    {
      theAccount = [[NSMutableDictionary alloc] initWithDictionary: [allAccounts objectForKey: aKey]];
      allValues = [[NSMutableDictionary alloc] initWithDictionary: [theAccount objectForKey: @"MAILBOXES"]];
      
      UPDATE_PATH(@"DRAFTSFOLDERNAME", theOldPath, thePath);
      UPDATE_PATH(@"INBOXFOLDERNAME", theOldPath, thePath);
      UPDATE_PATH(@"SENTFOLDERNAME", theOldPath, thePath);
      UPDATE_PATH(@"TRASHFOLDERNAME", theOldPath, thePath);
      
      [theAccount setObject: allValues  forKey: @"MAILBOXES"];
      RELEASE(allValues);
      
      [allAccounts setObject: theAccount  forKey: aKey];
      RELEASE(theAccount);
    }
  
  [[NSUserDefaults standardUserDefaults] setObject: allAccounts  forKey: @"ACCOUNTS"];
  RELEASE(allAccounts);
}


//
//
//
- (void) _updateShowOrHideReadAndDeletedForFolder: (Folder *) theFolder
{
  
  // We now set the folder to show or hide our deleted messages from the folder
  // Hiding them is now the default behaviour (in 1.1.0pre2)
  if ([[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_DELETED_MESSAGES"] &&
      [[[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_DELETED_MESSAGES"] intValue] == NSOffState)
    {
      [theFolder setShowDeleted: YES];
      [[NSApp delegate] updateShowOrHideDeletedMenuItem: YES];
    }
  else
    {
      [theFolder setShowDeleted: NO];
      [[NSApp delegate] updateShowOrHideDeletedMenuItem: NO];
    }

  // We now set the folder to show or hide read messages from the folder. Showing them
  // is the default behavior
  if (![[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_READ_MESSAGES"] ||
      [[[NSUserDefaults standardUserDefaults] objectForKey: @"HIDE_READ_MESSAGES"] intValue] == NSOffState)
    {
      [theFolder setShowRead: YES];
      [[NSApp delegate] updateShowOrHideReadMenuItem: YES];
    }
  else
    {
      [theFolder setShowRead: NO];
      [[NSApp delegate] updateShowOrHideReadMenuItem: NO];
    }
}

@end








