/*
**  NSAttributedString+Extensions.m
**
**  Copyright (c) 2004
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "NSAttributedString+Extensions.h"

#include "Constants.h"
#include "ExtendedTextAttachmentCell.h"
#include "FilterManager.h"
#include "GNUMail.h"
#include "MailHeaderCell.h"
#include "MailWindowController.h"
#include "MimeType.h"
#include "MimeTypeManager.h"
#include "NSAttributedString+TextEnriched.h"
#include "NSColor+Extensions.h"
#include "NSFont+Extensions.h"
#include "ThreadArcsCell.h"

#include <Pantomime/Folder.h>
#include <Pantomime/InternetAddress.h>
#include <Pantomime/Message.h>
#include <Pantomime/MimeMultipart.h>
#include <Pantomime/MimeUtility.h>
#include <Pantomime/NSString+Extensions.h>
#include <Pantomime/Part.h>

//
//
//
static inline int levelFromString(NSString *theString, int start, int end)
{
  int i, level;
  unichar c;
  
  for (i = start,level = 0; i < end; i++)
    {
      c = [theString characterAtIndex: i];

      if (c == '>')
	{
	  level++;
	}
      else if (c > 32)
	{
	  break;
	}
    }

  return level;
}

//
//
//
@implementation NSAttributedString (Extensions)

//
// This method returns a NSAttributedString object that has been built
// from the content of the message.
//
+ (NSAttributedString *) attributedStringFromContentForPart: (Part *) thePart
{
  NSMutableAttributedString *maStr;
  NSMutableDictionary *tAttr;
  
  tAttr = [[NSMutableDictionary alloc] init];
  
  [tAttr setObject: [NSFont messageFont]
	 forKey: NSFontAttributeName];
  
  maStr = [[NSMutableAttributedString alloc] init];
  
  if ( [[thePart content] isKindOfClass: [MimeMultipart class]] )
    {
      // We first verify if our multipart object is a multipart alternative.
      // If yes, we represent the best representation of the part.
      if ( [thePart isMimeType: @"multipart" : @"alternative"] )
	{
	  // We append our \n to separate our body part from the headers
	  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n" 
							     attributes: nil] ];
	  
	  // We then append the best representation from this multipart/alternative object
	  [maStr appendAttributedString: [NSAttributedString attributedStringFromMultipartAlternative:
							       (MimeMultipart *)[thePart content]] ];
	}
      // Then we verify if our multipart object is a multipart appledouble.
      else if ( [thePart isMimeType: @"multipart" : @"appledouble"] )
	{
	  // We append our \n to separate our body part from the headers
	  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n" 
							     attributes: nil] ];
	  
	  // We then append the best representation from this multipart/appledouble object
	  [maStr appendAttributedString: [NSAttributedString attributedStringFromMultipartAppleDouble:
							       (MimeMultipart *)[thePart content]] ];
	}
      // We have a multipart/mixed or multipart/related (or an other, unknown multipart/* object)
      else
	{
	  MimeMultipart *aMimeMultipart;
	  Part *aPart;
	  int i;
     
	  aMimeMultipart = (MimeMultipart*)[thePart content];
     
	  for (i = 0; i < [aMimeMultipart count]; i++)
	    {
	      // We get our part
	      aPart = [aMimeMultipart partAtIndex: i];
	      
	      // We recursively call our method to show all parts
	      [maStr appendAttributedString: [self attributedStringFromContentForPart: aPart]];
	    }
	}
    }
  // We have a message with Content-Type: application/* OR audio/* OR image/* OR video/*
  // We can also have a text/* part that was base64 encoded, but, we skip it. It'll be
  // treated as a NSString object. (See below).
  else if ([[thePart content] isKindOfClass: [NSData class]])
    {
      NSTextAttachment *aTextAttachment;
      ExtendedTextAttachmentCell *cell;
      NSFileWrapper *aFileWrapper;
      
      MimeType *aMimeType;
      NSImage *anImage;
      
      NSRect rectOfTextView;
      NSSize imageSize;
     
      aFileWrapper = [[NSFileWrapper alloc] initRegularFileWithContents: (NSData *)[thePart content]];
      
      if (![thePart filename])
	{
	  [aFileWrapper setPreferredFilename: @"unknown"];
	}
      else
	{
	  [aFileWrapper setPreferredFilename: [thePart filename]];
	}

      // We get the righ Mime-Type object for this part
      aMimeType = [[MimeTypeManager singleInstance] 
		    mimeTypeForFileExtension: [[aFileWrapper preferredFilename] pathExtension]];
      
      if (aMimeType && [aMimeType view] == DISPLAY_AS_ICON)
	{
	  anImage = [[MimeTypeManager singleInstance] bestIconForMimeType: aMimeType
						      pathExtension: [[aFileWrapper preferredFilename] pathExtension]];
	}
      else
	{
	  // If the image is a TIFF, we create it directly from a NSData.
	  if ([thePart isMimeType: @"image": @"*"])
	    {
	      anImage = [[NSImage alloc] initWithData:(NSData *)[thePart content]];
	      AUTORELEASE(anImage);
	    }
	  //
	  // It's not an image, we query our MIME type manager in order to see 
	  // to get the right icon.
	  //
	  else
	    {
	      anImage = [[MimeTypeManager singleInstance] bestIconForMimeType: aMimeType
							  pathExtension: [[aFileWrapper preferredFilename] 
									   pathExtension]];
	    }
	}
      
      // If the image has been loaded sucessfully, it become our icon for our filewrapper
      if (anImage)
	{
	  [aFileWrapper setIcon: anImage];
	} 	
      
      // We now rescale the attachment if it doesn't fit in the text view. That could happen
      // very often for images.
      rectOfTextView = [[[[GNUMail lastMailWindowOnTop] windowController] textView] frame];
      imageSize = [[aFileWrapper icon] size];
      
      if (imageSize.width > rectOfTextView.size.width)
	{
	  double delta =  1.0 / (imageSize.width / (rectOfTextView.size.width-10));
	  [[aFileWrapper icon] setScalesWhenResized: YES];
	  [[aFileWrapper icon] setSize: NSMakeSize((imageSize.width*delta), imageSize.height*delta)];
	}
      
      // We now create our text attachment with our file wrapper
      aTextAttachment = [[NSTextAttachment alloc] initWithFileWrapper: aFileWrapper];
      
      // We add this attachment to our 'Save Attachment' menu
      [(GNUMail *)[NSApp delegate] addItemToMenuFromTextAttachment: aTextAttachment];
        
      cell = [[ExtendedTextAttachmentCell alloc] initWithFilename: [aFileWrapper preferredFilename]
						 size: [(NSData *)[thePart content] length] ];
      [cell setPart: thePart];
      
      [aTextAttachment setAttachmentCell: cell];
      
      // Cocoa bug
#ifdef MACOSX
      [cell setAttachment: aTextAttachment];
      [cell setImage: [aFileWrapper icon]];
#endif
      RELEASE(cell);
      RELEASE(aFileWrapper);
      
      // We separate the text attachment from any other line previously shown
      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n" 
							 attributes: nil] ];
      
      [maStr appendAttributedString: [NSAttributedString attributedStringWithAttachment: aTextAttachment]];
      RELEASE(aTextAttachment);
    }
  //
  // We have a message/rfc822 as the Content-Type.
  //
  else if ( [[thePart content] isKindOfClass: [Message class]] )
    {
      Message *aMessage;
      
      aMessage = (Message *)[thePart content];
      
      // We must represent this message/rfc822 part as an attachment
      if ( [thePart contentDisposition] &&
	   ([[thePart contentDisposition] caseInsensitiveCompare: @"attachment"] == NSOrderedSame) )
	{
	  NSTextAttachment *aTextAttachment;
	  ExtendedTextAttachmentCell *cell;
	  NSFileWrapper *aFileWrapper;
	  NSData *aData;
	  
	  aData = [aMessage rawSource];
	  
	  aFileWrapper = [[NSFileWrapper alloc] initRegularFileWithContents: aData];
	  
	  [aFileWrapper setPreferredFilename: @"message/rfc822 attachment"];
	  
	  aTextAttachment = [[NSTextAttachment alloc] initWithFileWrapper: aFileWrapper];
	  
	  // We add this attachment to our 'Save Attachment' menu
	  [(GNUMail *)[NSApp delegate] addItemToMenuFromTextAttachment: aTextAttachment];
          
	  cell = [[ExtendedTextAttachmentCell alloc] initWithFilename: [aFileWrapper preferredFilename]
						     size: [aData length] ];
	  [cell setPart: thePart];
	  
	  [aTextAttachment setAttachmentCell: cell];
          
          // Cocoa bug
#ifdef MACOSX
          [cell setAttachment: aTextAttachment];
          [cell setImage: [aFileWrapper icon]];
#endif
	  RELEASE(cell); 
	  RELEASE(aFileWrapper);
	  
	  [maStr appendAttributedString: [NSAttributedString attributedStringWithAttachment:
							       aTextAttachment]];
	  RELEASE(aTextAttachment);
	}
      // Its inline..
      else
	{	      
	  [maStr appendAttributedString: [self attributedStringFromHeadersForMessage: aMessage
					       showAllHeaders: NO
					       useMailHeaderCell: NO] ];
	  [maStr appendAttributedString: [self attributedStringFromContentForPart: aMessage]];
	}
    }
  //
  // We have a message with Content-Type: text/*
  // or at least, a part that we can represent directly in our textView.
  // 
  else if ( [[thePart content] isKindOfClass: [NSString class]] )
    { 
      NSString *aString;
      
      aString = (NSString *)[thePart content];

      // We must represent our NSString as an text attachment in our textView.
      if ( [thePart contentDisposition] &&
	   ([[thePart contentDisposition] caseInsensitiveCompare: @"attachment"] == NSOrderedSame) )
	{
	  NSTextAttachment *aTextAttachment;
	  ExtendedTextAttachmentCell *cell;
	  NSFileWrapper *aFileWrapper;
	  NSData *aData;
	  
	  aData = [aString dataUsingEncoding: NSUTF8StringEncoding];
	  aFileWrapper = [[NSFileWrapper alloc] initRegularFileWithContents: aData];
	  
	  if ( ![thePart filename] )
	    {
	      [aFileWrapper setPreferredFilename: @"unknown"];
	    }
	  else
	    {
	      [aFileWrapper setPreferredFilename: [thePart filename]];
	    }
	  
	  aTextAttachment = [[NSTextAttachment alloc] initWithFileWrapper: aFileWrapper];
	  
	  // We add this attachment to our 'Save Attachment' menu
	  [(GNUMail *)[NSApp delegate] addItemToMenuFromTextAttachment: aTextAttachment];

	  cell = [[ExtendedTextAttachmentCell alloc] initWithFilename: [aFileWrapper preferredFilename]
						     size: [aData length] ];
	  [cell setPart: thePart];
	  [aTextAttachment setAttachmentCell: cell];
          
          // Cocoa bug
#ifdef MACOSX
          [cell setAttachment: aTextAttachment];
          [cell setImage: [aFileWrapper icon]];
#endif
	  RELEASE(cell);
	  RELEASE(aFileWrapper);
	  
	  // We separate the text attachment from any other line previously shown
	  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n" 
							     attributes: nil] ];
	  
	  [maStr appendAttributedString: [NSAttributedString attributedStringWithAttachment:
							       aTextAttachment]];
	  RELEASE(aTextAttachment);
	}
      // It's inline...
      else
	{
	  [maStr appendAttributedString: [NSAttributedString attributedStringFromTextPart: thePart]];
	}
    }
  //
  // We have something that we probably can't display.
  // Let's inform the user about this situation.
  // 
  else
    {
      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: _(@"Loading message...")
							 attributes: nil]];
    }
  
  RELEASE(tAttr);

  return AUTORELEASE(maStr);
}


//
// This method returns a NSAttributedString object that has been built
// from the headers of the message.
//
+ (NSAttributedString *) attributedStringFromHeadersForMessage: (Message *) theMessage
						showAllHeaders: (BOOL) showAllHeaders
					     useMailHeaderCell: (BOOL) useMailHeaderCell
{
  NSMutableDictionary *headerNameAttribute, *headerValueAttribute;
  NSMutableAttributedString *maStr; 

  NSMutableAttributedString *aMutableAttributedString;

  NSDictionary *allHeaders;
  NSArray *headersToShow;
  int i;

  maStr = [[NSMutableAttributedString alloc] init];
  

  // Attributes for our header names
  headerNameAttribute = [[NSMutableDictionary alloc] init];

  [headerNameAttribute setObject: [NSFont headerNameFont]
		       forKey: NSFontAttributeName];


  // Attributes for our header values
  headerValueAttribute = [[NSMutableDictionary alloc] init];

  [headerValueAttribute setObject: [NSFont headerValueFont]
			forKey: NSFontAttributeName];
			
  
  // We get all the message's headers
  allHeaders = [theMessage allHeaders];
  
  // We verify which headers of the message we show show.
  if ( showAllHeaders )
    {
      headersToShow = [allHeaders allKeys];
    }
  else
    {
      headersToShow = [[NSUserDefaults standardUserDefaults] objectForKey: @"SHOWNHEADERS"];
    }

  if ( headersToShow )
    {  
      for (i = 0; i < [headersToShow count]; i++)
	{
	  NSString *anHeader = [headersToShow objectAtIndex: i];
	  
	  if ( [anHeader caseInsensitiveCompare: @"Date"] == NSOrderedSame &&
	       [theMessage receivedDate] )
	    {
	      if ( [theMessage receivedDate] )
		{
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"Date: "
								     attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: [[theMessage receivedDate] description]
								     attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								     attributes: nil] ];
		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"From"] == NSOrderedSame &&
		    [theMessage from] )
	    {
	      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"From: "
								 attributes: headerNameAttribute] ];
	      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: [[theMessage from] stringValue]
								 attributes: headerValueAttribute] ];
	      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								 attributes: nil] ];
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Bcc"] == NSOrderedSame )
	    {
	      NSString *bccStr = [NSString stringFromRecipients: [theMessage recipients]
					   type: PantomimeBccRecipient];
	      
	      if ( [bccStr length] > 0 )
		{
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"Bcc: "
								     attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: [bccStr substringToIndex: ([bccStr length] - 2)]
								     attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								     attributes: nil] ];
		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Cc"] == NSOrderedSame )
	    {
	      NSString *ccStr= [NSString stringFromRecipients: [theMessage recipients]
					 type: PantomimeCcRecipient];
	      
	      if ( [ccStr length] > 0 )
		{
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"Cc: "
								     attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: [ccStr substringToIndex: ([ccStr length] - 2)]
								     attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								     attributes: nil] ];		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Reply-To"] == NSOrderedSame )
	    { 
	      InternetAddress *anInternetAddress = [theMessage replyTo];
	      
	      if ( anInternetAddress )
		{
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"Reply-To: "
								     attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: [[theMessage replyTo] stringValue]
								     attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								     attributes: nil] ];
		}	
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"To"] == NSOrderedSame )
	    {
	      NSString *toStr = [NSString stringFromRecipients: [theMessage recipients]
					  type: PantomimeToRecipient];
	      
	      if ( [toStr length] > 0 )
		{
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"To: "
								     attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: [toStr substringToIndex: ([toStr length] - 2)]
								     attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								     attributes: nil] ];
		}
	    }
	  else if ( [anHeader caseInsensitiveCompare: @"Content-Type"] == NSOrderedSame )
	    {
	      NSString *aString;
	      
	      if ( [theMessage charset] )
		{
		  aString = [NSString stringWithFormat: @"%@; charset=%@",
				      [theMessage contentType], [theMessage charset]];
		}
	      else
		{
		  aString = [theMessage contentType];
		}
	      
	      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"Content-Type: "
								 attributes: headerNameAttribute] ];
	      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: aString
								 attributes: headerValueAttribute] ];
	      [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								 attributes: nil] ];
	    }
	  else
	    {
	      NSArray *allKeys = [allHeaders allKeys];
	      NSString *valueString = nil;
	      int j, count;
	      
	      count = [allKeys count];
	      
	      for (j = 0; j < count; j++)
		{
		  if ( [[[allKeys objectAtIndex: j] uppercaseString] isEqualToString: [anHeader uppercaseString]] )
		    {
		      if ( [[allHeaders objectForKey: [allKeys objectAtIndex: j]] 
			     isKindOfClass: [InternetAddress class]] )
			{
			  valueString = [[allHeaders objectForKey: [allKeys objectAtIndex: j]] stringValue];
			}
		      else
			{
			  valueString = [allHeaders objectForKey: [allKeys objectAtIndex: j]];
			}
		      break;
		    }
		}
		  
	      if ( valueString )
		{
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: [NSString stringWithFormat: @"%@: ", anHeader]
								     attributes: headerNameAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: valueString
								     attributes: headerValueAttribute] ];
		  [maStr appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
								     attributes: nil] ];
		}
	    }
	} // for (..) 
      
    } // if ( headersToShow )
  
  if ( useMailHeaderCell )
    {
      NSTextAttachment *aTextAttachment;
      MailHeaderCell *theMailHeaderCell;
      
      theMailHeaderCell = [[[GNUMail lastMailWindowOnTop] windowController] mailHeaderCell];
      [theMailHeaderCell setColor: [[FilterManager singleInstance] colorForMessage: theMessage]];
      [theMailHeaderCell setAttributedStringValue: maStr];
      [theMailHeaderCell resize: nil];
      
      // We now "embed" the header cell into a NSTextAttachment object
      // so we can add it to our mutable attributed string.
      aTextAttachment = [[NSTextAttachment alloc] init];
      [aTextAttachment setAttachmentCell: theMailHeaderCell];
      
      aMutableAttributedString = [[NSMutableAttributedString alloc] init];
      
      [aMutableAttributedString appendAttributedString: 
				  [NSMutableAttributedString attributedStringWithAttachment: aTextAttachment]];
      RELEASE(aTextAttachment);
      
      //
      // If we are using message thread, we show our visualization cell.
      //
      if ([[theMessage folder] allContainers])
	{
	  aTextAttachment = [[NSTextAttachment alloc] init];
	  [aTextAttachment setAttachmentCell: AUTORELEASE([[ThreadArcsCell alloc] init])];
	  [aMutableAttributedString appendAttributedString: [NSMutableAttributedString attributedStringWithAttachment: aTextAttachment]];
	  RELEASE(aTextAttachment);
	}
      [aMutableAttributedString appendAttributedString: [NSAttributedString attributedStringWithString: @"\n\n"
									    attributes: nil]];
    }
  else
    {
      aMutableAttributedString = [[NSMutableAttributedString alloc] init];
      [aMutableAttributedString appendAttributedString: maStr];
      [aMutableAttributedString appendAttributedString: [NSAttributedString attributedStringWithString: @"\n"
									    attributes: nil] ];
    }
  
  RELEASE(maStr);
  RELEASE(headerNameAttribute);
  RELEASE(headerValueAttribute);

  return AUTORELEASE(aMutableAttributedString);
}


//
//
//
+ (NSAttributedString *) attributedStringFromMultipartAlternative: (MimeMultipart *) theMimeMultipart
{
  NSString *aSubtype;
  Part *aPart;
  int i, index;
  
  if ([[NSUserDefaults standardUserDefaults] integerForKey: @"DEFAULT_MULTIPART_ALTERNATIVE_TYPE"] == TYPE_HTML)
    {
      aSubtype = @"html";
    }
  else
    {
      aSubtype = @"plain";
    }
  
  index = -1;

  // We search for our preferred part (text/html or text/plain) depending on aSubtype
  for (i = 0; i < [theMimeMultipart count]; i++)
    {
      aPart = [theMimeMultipart partAtIndex: i];
      
      if ([aPart isMimeType: @"text": aSubtype])
	{
	  index = i;
	  break;
	}
    }

  // If we have found our preferred part, we use that one
  if (index >= 0)
    {
      return [self attributedStringFromTextPart: [theMimeMultipart partAtIndex: index]];
    }
  // We haven't, we show the first one!
  else
    {
      if ([theMimeMultipart count] > 0)
	{
	  return [self attributedStringFromTextPart: [theMimeMultipart partAtIndex: 0]];
	}
    }
  
  return [self attributedStringFromTextPart: nil];
}


//
//
//
+ (NSAttributedString *) attributedStringFromMultipartAppleDouble: (MimeMultipart *) theMimeMultipart
{
  NSMutableAttributedString *aMutableAttributedString;
  NSMutableDictionary *attributes;
  Part *aPart;
  int i;
  
  // We create a set of attributes (base font, color red)
  attributes = [[NSMutableDictionary alloc] init];
  [attributes setObject: [NSColor redColor]
	      forKey: NSForegroundColorAttributeName];

  aMutableAttributedString = [[NSMutableAttributedString alloc] init];

  for (i = 0; i < [theMimeMultipart count]; i++)
    {
      aPart = [theMimeMultipart partAtIndex: i];
    
      if ( [aPart isMimeType: @"application": @"applefile"] )
	{
	  [aMutableAttributedString appendAttributedString: [self attributedStringWithString: _(@"(Decoded Apple file follows...)")
								  attributes: attributes]];
	}
      else
	{
	  // We first add a \n between our applefile description and the 'representation' of the attachment
	  [aMutableAttributedString appendAttributedString: [self attributedStringWithString: @"\n"  attributes: nil]];
	  
	  // We add the representation of our attachment
	  [aMutableAttributedString appendAttributedString: [self attributedStringFromContentForPart: aPart]];
	} 
    }
  
  // We add a \n to separate everything.
  [aMutableAttributedString appendAttributedString:
			      [NSAttributedString attributedStringWithString: @"\n" attributes: nil]];
  
  RELEASE(attributes);

  return AUTORELEASE(aMutableAttributedString);
}


//
//
//
+ (NSAttributedString *) attributedStringFromTextPart: (Part *) thePart
{
  NSMutableDictionary *textMessageAttributes;
  NSAttributedString *aAttributedString;
  NSString *aString;

  aAttributedString = nil;

  // Initial and trivial verifications.
  // We MUST be sure the content isn't nil and also that it's a valid NSString object.
  // Pantomime will return the raw source (NSData object) if the decoding operation has failed.
  if ( !thePart ||
       ![thePart content] || 
       ![[thePart content] isKindOfClass: [NSString class]] )
    {
      goto done;
    }

  // Initializations of the local vars
  textMessageAttributes = [[NSMutableDictionary alloc] init];
  aString = (NSString *)[thePart content];
  
  [textMessageAttributes setObject: [NSFont messageFont]
			 forKey: NSFontAttributeName];


  //
  // text/html
  //
  if ( [thePart isMimeType: @"text": @"html"] )
    {
#ifdef MACOSX
      NSData *aData;
      
      aData = [aString dataUsingEncoding: [NSString encodingForPart: thePart]];      
      aAttributedString = [[NSAttributedString alloc] initWithHTML: aData
						      documentAttributes: nil];
      AUTORELEASE(aAttributedString);
#else
      aString = [MimeUtility plainTextContentFromPart: thePart];
      aAttributedString = [NSAttributedString attributedStringWithString: aString
					      attributes: textMessageAttributes];
#endif
    }
  //
  // text/enriched
  //
  else if ( [thePart isMimeType: @"text": @"enriched"] )
    {
      aAttributedString = [NSAttributedString attributedStringFromTextEnrichedString: aString];
    }
  //
  // text/rtf
  //
  else if ( [thePart isMimeType: @"text": @"rtf"] )
    {
      NSData *aData;
	      
      aData = [aString dataUsingEncoding: [NSString encodingForPart: thePart]];
      
      aAttributedString = [[NSAttributedString alloc] initWithRTF: aData
						      documentAttributes: NULL];
      AUTORELEASE(aAttributedString); 
    }
  //
  // We surely got a text/plain part
  //
  else
    {
      NSMutableDictionary *plainTextMessageAttributes;
      
      if ( [[NSUserDefaults standardUserDefaults] 
	     objectForKey: @"USE_FIXED_FONT_FOR_TEXT_PLAIN_MESSAGES"] &&
	   [[NSUserDefaults standardUserDefaults] 
	     integerForKey: @"USE_FIXED_FONT_FOR_TEXT_PLAIN_MESSAGES"] == NSOnState )
	{
	  plainTextMessageAttributes = [[NSMutableDictionary alloc] init];
	  AUTORELEASE(plainTextMessageAttributes);
	  
	  [plainTextMessageAttributes setObject: [NSFont plainTextMessageFont]
				      forKey: NSFontAttributeName];
	}
      else
	{
	  plainTextMessageAttributes = textMessageAttributes;
	}
      
      aAttributedString = [NSAttributedString attributedStringWithString: aString
					      attributes: plainTextMessageAttributes];
    }

  RELEASE(textMessageAttributes);
  
 done:
  // We haven't found a text/plain part. Report this as a bug in GNUMail.app for not supporting
  // the other kind of parts.
  if ( !aAttributedString )
    {
      aAttributedString = [NSAttributedString attributedStringWithString: 
						_(@"No text part found. Please report this bug since GNUMail.app doesn't support this kind of part.")
					      attributes: nil];
    }
  
  return aAttributedString;
}


//
//
//
+ (NSAttributedString *) attributedStringWithString: (NSString *) theString
					 attributes: (NSDictionary *) theAttributes
{
  if (!theAttributes)
    {
      NSMutableDictionary *aMutableAttributedString;
      NSMutableDictionary *attributes;

      attributes = [[NSMutableDictionary alloc] init];
      [attributes setObject: [NSFont systemFontOfSize: 0] forKey: NSFontAttributeName];
      aMutableAttributedString = [[self alloc] initWithString: theString   attributes: attributes];
      RELEASE(attributes);
      
      return AUTORELEASE(aMutableAttributedString);
    }
  
  
  return AUTORELEASE([[NSAttributedString alloc] initWithString: theString attributes: theAttributes]);
}


//
//
//
#warning implement and use in the cell drawing code for the message headers
- (NSSize) sizeInRect: (NSRect) theRect
{
  if ([self size].width < theRect.size.width)
    {
      return [self size];
    }

  return NSZeroSize;
}

@end


//
//
//
@implementation NSMutableAttributedString (Extensions)

- (void) format
{
  NSMutableAttributedString *aMutableAttributedString;
  NSRange aRange;
  
  id attribute;
  int index;
  
  //
  // If we only have one char (or less), we don't do anything.
  // This could append for example when we only have an image/tiff as the Content-Type of an email.
  //
  if ([self length] <= 1 )
    {
      return;
    }
  
  aMutableAttributedString = nil;
  index = 0;

  attribute = [self attribute: NSAttachmentAttributeName
		    atIndex: index
		    effectiveRange: &aRange];
  
  //
  // We if have NO text attachments at all, we simply return.
  //
  if (!attribute && aRange.length == [self length])
    {
      return;
    }
  
  while (index < [self length])
    {
      // We found a text attachment...
      if ( attribute )
	{
	  // We search for <<filename>> OR <filename> and if we got one of them,
	  // we simply replace that string with the actual attachment.
	  ExtendedTextAttachmentCell *cell;
	  
	  cell = (ExtendedTextAttachmentCell *)[attribute attachmentCell];

	  if ([cell respondsToSelector: @selector(part)])
	    {
	      NSString *aString;
	      NSRange r;
	      aString = [self string];
	      
	      // We first search for <<filename>>
	      r = [aString rangeOfString: [NSString stringWithFormat: @"<<%@>>", [[cell part] filename]]];
	      
	      // If not found, we search for <filename>
	      if ( r.length == 0 )
		{
		  r = [aString rangeOfString: [NSString stringWithFormat: @"<%@>", [[cell part] filename]]];
		}
	      
	      // We found something, let's replace <filename> (or <<filename>>) by the actual
	      // attachment cell.
	      if ( r.length )
		{
		  NSAttributedString *attachment;
		  
		  // We get the attributed that contains our attachment
		  attachment = [self attributedSubstringFromRange: aRange];
		  RETAIN(attachment);
	      
		  // FIXME - we should verify if the range of that attachment > of the range of
		  // the <filename in case the disposition of our E-Mail is 'wrong'.
		  
		  // We remove it from our string
		  [self deleteCharactersInRange: aRange];
		  
		  // We replace
		  [self replaceCharactersInRange: r  withAttributedString: attachment];
		  
		  RELEASE(attachment);
		  
		  // We adjust our index for the search
		  index = r.location + 1;
		}
	      else
		{
		  index = NSMaxRange(r);
		}
	    }
	  else
	    {
	      index = NSMaxRange(aRange);
	    }
	}
      else // of if ( attribute ) 
	{
	  index = NSMaxRange(aRange);
	}
      
      if (index >= [self length])
      	{
      	  break;
      	}

      attribute = [self attribute: NSAttachmentAttributeName
			atIndex: index
			effectiveRange: &aRange];
    }
}


//
//
//
- (void) highlightAndActivateURL
{
  NSRange searchRange, foundRange;
  NSString *aString, *aPrefix;
  NSEnumerator *theEnumerator;
  NSArray *allPrefixes;

  int len;
  char c;

  allPrefixes = [NSArray arrayWithObjects: @"www.", @"http://", @"https://", @"ftp://", @"file://", nil];
  theEnumerator = [allPrefixes objectEnumerator];
    
  aString = [self string];
  len = [aString length];

  while ( (aPrefix = (NSString *)[theEnumerator nextObject]) )
    {
      searchRange = NSMakeRange(0, len);
      do
	{
	  foundRange = [aString rangeOfString: aPrefix
				options: 0
				range: searchRange];
	  
	  // If we found an URL...
	  if ( foundRange.length > 0 )
	    {
	      NSDictionary *linkAttributes;
	      NSURL *anURL;
	      int end;

	      // Restrict the searchRange so that it won't find the same string again
	      searchRange.location = end = NSMaxRange(foundRange);
	      searchRange.length = len - searchRange.location;
	      
	      // We assume the URL ends with whitespace
	      while ( (end < len) && ((c = [aString characterAtIndex: end]) != '\n' && c != ' ' && c != '\t'))
		{
		  end++;
		}

	      // Set foundRange's length to the length of the URL
	      foundRange.length = end - foundRange.location;

	      // If our URL is ended with a ".", "!", "," or ">", we trim it
	      c = [aString characterAtIndex: (end - 1)];
	      if ( c == '.' || c == '!' || c == ',' || c == '?' || c == '>')
		{
		  foundRange.length--;
		}

	      // We create our URL object from our substring
              // if we found just "www", we prepend "http://"
              if ( [aPrefix caseInsensitiveCompare: @"www."] == NSOrderedSame )
		{
		  anURL = [NSURL URLWithString: [NSString stringWithFormat: @"http://%@", 
							  [aString substringWithRange: foundRange]]];
		}
	      else
		{
		  anURL = [NSURL URLWithString: [aString substringWithRange: foundRange]];
		}
	      
	      // Make the link attributes
	      linkAttributes= [NSDictionary dictionaryWithObjectsAndKeys: anURL, NSLinkAttributeName,
					    [NSNumber numberWithInt: NSSingleUnderlineStyle], NSUnderlineStyleAttributeName,
					    [NSColor blueColor], NSForegroundColorAttributeName,
					    NULL];
	      
	      // Finally, apply those attributes to the URL in the text
	      [self addAttributes: linkAttributes  range: foundRange];
	    }
	  
	} while ( foundRange.length != 0 );
    }
}


//
//
//
- (void) quote
{
  if ([[NSUserDefaults standardUserDefaults] integerForKey: @"COLOR_QUOTED_TEXT"] == NSOffState)
    {
      return;
    }
  else
    {   
      NSDictionary *attributes;
      NSString *aString;
 
      int i, j, len, level;
     
      aString = [self string];
      len = [aString length];
      i = j = level = 0;
      
      for (; i < len; i++)
	{
	  if ( [aString characterAtIndex: i]=='\n' )
	    {
	      if (i > j)
		{
		  level = levelFromString(aString, j, i);
		  
		  if (level)
		    {
		      attributes = [NSDictionary dictionaryWithObjectsAndKeys: 
						   [NSColor colorForLevel: level],
						 NSForegroundColorAttributeName, 
						 nil];
		      [self addAttributes: attributes  range: NSMakeRange(j,i-j)];
		    }
		}
	      j = i+1;
	    }
	}
      
      if (i > j)
	{
	  level = levelFromString(aString, j, i);
	  if (level)
	    {
	      attributes = [NSDictionary dictionaryWithObjectsAndKeys: 
					   [NSColor colorForLevel: level],
					 NSForegroundColorAttributeName, 
					 nil];
	      [self addAttributes: attributes  range: NSMakeRange(j,i-j)];
	    }
	}
    }
}

@end

