/*
**  TaskManager.m
**
**  Copyright (c) 2002-2004
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "TaskManager.h"

#include "ConsoleWindowController.h"
#include "Constants.h"
#include "ExtendedTableView.h"
#include "Filter.h"
#include "FilterManager.h"
#include "FindWindowController.h"
#include "GNUMail.h"
#include "LocalMailDelivery.h"
#include "MailboxManagerController.h"
#include "MailWindowController.h"
#include "MessageViewWindowController.h"
#include "Task.h"
#include "Utilities.h"

#include <Foundation/NSArray.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSData.h>
#include <Foundation/NSDate.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSTimer.h>
#include <Foundation/NSValue.h>

#include <Pantomime/Constants.h>
#include <Pantomime/FolderInformation.h>
#include <Pantomime/IMAPStore.h>
#include <Pantomime/Message.h>
#include <Pantomime/POP3CacheManager.h>
#include <Pantomime/POP3CacheObject.h>
#include <Pantomime/POP3Folder.h>
#include <Pantomime/POP3Message.h>
#include <Pantomime/POP3Store.h>
#include <Pantomime/Sendmail.h>
#include <Pantomime/SMTP.h>
#include <Pantomime/TCPConnection.h>
#include <Pantomime/URLName.h>

#include <unistd.h>

static TaskManager *singleInstance;
static Class IMAPStore_class;
static Class LocalStore_class;
static Class POP3Store_class;
static Class SMTP_class;


//
// TaskManager Private Interface
//
@interface TaskManager (Private)

- (void) _checkMailForAccount: (NSString *) theAccountName
		       origin: (int) theOrigin
			owner: (id) theOwner;

- (void) _executeActionUsingFilter: (Filter *) theFilter
                           message: (NSData *) theMessage
                              task: (Task *) theTask;

- (BOOL) _matchFilterRuleFromRawSource: (NSData *) theRawSource
                                  task: (Task *) theTask;

- (void) _nextTask;

- (void) _receiveUsingIMAPForTask: (Task *) theTask;
- (void) _receiveUsingPOP3ForTask: (Task *) theTask;
- (void) _receiveUsingUNIXForTask: (Task *) theTask;

- (void) _sendUsingSendmailForTask: (Task *) theTask;
- (void) _sendUsingSMTPForTask: (Task *) theTask;
- (void) _tick;
- (void) _tick_internal;

@end



//
//
//
@implementation TaskManager

+ (void) initialize
{
  IMAPStore_class = [IMAPStore class];
  LocalStore_class = [LocalStore class];
  POP3Store_class = [POP3Store class];
  SMTP_class = [SMTP class];
}


//
//
//
- (id) init
{
  self = [super init];

  _table = NSCreateMapTable(NSObjectMapKeyCallBacks, NSObjectMapValueCallBacks, 64);
  _tasks = [[NSMutableArray alloc] init];
  _counter = 0;

  return self;
}


//
//
//
- (void) dealloc
{
  NSDebugLog(@"TaskManager: -dealloc");

  NSFreeMapTable(_table);
  RELEASE(_tasks);

  [super dealloc];
}


//
// This method is used to add a task to the runqueue. It will NOT
// add a task that is identical to one already in the runqueue.
//
// Those tasks include: RECEIVE_POP3 and RECEIVE_IMAP+IMAP_STATUS
//
- (void) addTask: (Task *) theTask
{
  // We insert it at the beginning of our array
  if (theTask)
    {
      Task *aTask;
      int count;

      count = [_tasks count];

      while (count--)
	{
	  aTask = [_tasks objectAtIndex: count];

	  if ([theTask op] != SEND_SMTP && 
	      [theTask op] != SEND_SENDMAIL &&
	      [[aTask key] isEqualToString: [theTask key]])
	    {
	      return;
	    }
	}
      
      [_tasks insertObject: theTask  atIndex: 0];
      [self _nextTask];
      
      // We start our animations and reload the Console
      count = [[GNUMail allMailWindows] count];

      while (count--)
	{
	  [[[[GNUMail allMailWindows] objectAtIndex: count] windowController] startAnimation];
	}

      [[[ConsoleWindowController singleInstance] tasksTableView] reloadData];
    }
}


//
//
//
- (void) removeTask: (Task *) theTask
{
  [_tasks removeObject: theTask];
  [[[ConsoleWindowController singleInstance] tasksTableView] reloadData];
  [[ConsoleWindowController singleInstance] updateToggle];
}


//
//
//
- (NSArray *) allTasks
{
  return [NSArray arrayWithArray: _tasks];
}


//
//
//
- (void) checkForNewMail: (id) theSender
	      controller: (MailWindowController *) theMailWindowController
{
  NSArray *allKeys;
  int i, origin;

  // First of all, we get the 'origin' of the action. That is, if it was 
  // initiated from the user, from the timer or from self, when verifying
  // mail on statup.
  if (theSender == theMailWindowController ||
      theSender == [NSApp delegate])
    {
      origin = ORIGIN_STARTUP;
    }
  else
    {
      // We clicked on the get button or on the menu item.
      origin = ORIGIN_USER;
    }

  // We reset our counter ivar since we don't want to check mails immediately
  // after we already checked for it.
  _counter = 0;
  
  //
  // If it's our menu item that called this method but, it's NOT the get "All"
  // that has been clicked. So, we must verify only one account.
  //
  if (theSender && 
      [theSender isKindOfClass: [NSMenuItem class]] &&
      [theSender tag] >= 0)
    {
      allKeys = [NSArray arrayWithObject: [theSender title]];
    }
  //
  // If the user has clicked on the Get button OR 
  // It is our "All" menu item that was clicked on.
  //
  else if ((theMailWindowController && theSender == theMailWindowController->get) ||
	   (theSender && [theSender isKindOfClass: [NSMenuItem class]] && [theSender tag] < 0))
    {
      // We get all accounts here but we'll verify later that we only use POP3 accounts.
      allKeys = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] allKeys]
		  sortedArrayUsingSelector: @selector(compare:)];
    }
  //
  // Our sender is the app's delegate. That means that we were asked to verify mails 
  // on startup. Let's only get the POP3 and UNIX accounts that we must verify on startup.
  // We skip over IMAP accounts since we don't want to establish the required connections.
  //
  else if (theSender == [NSApp delegate])
    {
      NSMutableArray *aMutableArray;
      
      aMutableArray = [[NSMutableArray alloc] initWithArray: [[[NSUserDefaults standardUserDefaults]
								objectForKey: @"ACCOUNTS"] allKeys]];      
      for (i = ([aMutableArray count]-1); i >= 0; i--)
	{
	  NSDictionary *allValues;
	  NSString *aKey;
	  
	  aKey = [aMutableArray objectAtIndex: i];

	  // If the account is disabled or it's set to never check mail, we skip over it
	  if ( ![[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] objectForKey: aKey]
		   objectForKey: @"ENABLED"] boolValue] ||
	       [[[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] objectForKey: aKey]
		   objectForKey: @"RECEIVE"] objectForKey: @"RETRIEVEMETHOD"] intValue] == NEVER )
	    {
	      continue;
	    }

	  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
			 objectForKey: aKey] objectForKey: @"RECEIVE"];
	  
	  if ( ![allValues objectForKey: @"CHECKONSTARTUP"] ||
	       [[allValues objectForKey: @"CHECKONSTARTUP"] intValue] == NSOffState ||
	       ([allValues objectForKey: @"SERVERTYPE"] && [[allValues objectForKey: @"SERVERTYPE"] intValue] == IMAP) )
	    {
	      [aMutableArray removeObject: aKey];
	    }
	  else
	    {
	      NSDebugLog(@"Will verify for new mail on startup for %@", aKey);
	    }
	}

      allKeys = AUTORELEASE(aMutableArray);
    }
  else
    {
      NSDebugLog(@"TaskManager: -checkForNewMail: controller: called w/o being handled.");
      return;
    }
  
  
  // If we got no keys, we immediately stop our animation
  if ([allKeys count] == 0 && theMailWindowController)
    {
      [theMailWindowController stopAnimation];
    }

  // We send all our tasks!
  for (i = 0; i < [allKeys count]; i++)
    {
      [self _checkMailForAccount: [allKeys objectAtIndex: i]
	    origin: origin
	    owner: theMailWindowController];
    }
}


//
//
//
- (void) taskCompleted: (Task *) theTask
{
  int i;

  if (theTask)
    {
      // If it was a RECEIVE task and we transferred messages to other mailboxes,
      // we warn the user about this.
      if ([theTask op] == RECEIVE_POP3 || [theTask op] == RECEIVE_UNIX)
	{
	  if ( [theTask receivedMessagesCount] > 0 &&
	       ([theTask origin] == ORIGIN_STARTUP || [theTask origin] == ORIGIN_TIMER) )
	    {
	      // We verify if we must play a sound. If the sound isn't found, we simply use NSBeep()
	      if ( [[NSUserDefaults standardUserDefaults] boolForKey: @"PLAY_SOUND"] )
		{
		  NSString *aString;

		  aString = [[NSUserDefaults standardUserDefaults] stringForKey: @"PATH_TO_SOUND"];
		  if ([[NSFileManager defaultManager] fileExistsAtPath: aString])
		    {
		      NSSound *aSound;
		      
		      aSound = [[NSSound alloc] initWithContentsOfFile: aString  byReference: YES];
		      [aSound play];
		      RELEASE(aSound);
		    }
		  else
		    {
		      NSBeep();
		    }
		}
	    }
	  
	  if ([[theTask filteredMessagesFolders] count] > 0)
	    {
	      if ( [[NSUserDefaults standardUserDefaults] boolForKey: @"SHOW_FILTER_PANEL"] )
		{
		  NSRunInformationalAlertPanel(_(@"Filtered messages..."),
					       _(@"%d messages have been transfered to the following folders:\n%@"),
					       _(@"OK"),
					       NULL,
					       NULL,
					       [theTask filteredMessagesCount],
					       [[theTask filteredMessagesFolders] componentsJoinedByString: @"\n"]);
		}

	      if ([[NSUserDefaults standardUserDefaults] boolForKey: @"OPEN_MAILBOX_AFTER_TRANSFER"])
		{
		  NSString *aString, *aStoreName, *aFolderName;
		  URLName *theURLName;
		  NSRange aRange;
 
		  for (i = 0; i < [[theTask filteredMessagesFolders] count]; i++)
		    {
		      aString = [[theTask filteredMessagesFolders] objectAtIndex: i];
		      aRange = [aString rangeOfString: @" - "];
		      
		      aFolderName = [aString substringFromIndex: NSMaxRange(aRange)];
		      aStoreName = [aString substringToIndex: aRange.location];

		      if ([aStoreName isEqualToString: _(@"Local")])
			{
			  NSString *thePathToLocalMailDir;

			  thePathToLocalMailDir = [[NSUserDefaults standardUserDefaults] objectForKey: @"LOCALMAILDIR"];
			  theURLName = [[URLName alloc] initWithString: [NSString stringWithFormat: @"local://%@/%@", 
										  thePathToLocalMailDir, aFolderName]
							path: thePathToLocalMailDir];
			  
			}
		      else
			{
			  // We skip the "IMAP " part
			  aRange = [aStoreName rangeOfString: _(@"IMAP ")];
			  aStoreName = [aStoreName substringFromIndex: NSMaxRange(aRange)];
			  theURLName = [[URLName alloc] initWithString: [NSString stringWithFormat: @"imap://%@/%@",
										  aStoreName, aFolderName]];
			}
		      
		      // We finally opens the folder and release our URLName object.
		      [[MailboxManagerController singleInstance] openFolderWithURLName: theURLName
								 sender: [NSApp delegate]];
		      RELEASE(theURLName);
		    }    
		}
	    }
	}

      [self removeTask: theTask];

      if ([_tasks count] == 0)
	{
	  int count;

	  count = [[GNUMail allMailWindows] count];
	  
	  while (count--)
	    {
	      [[[[GNUMail allMailWindows] objectAtIndex: count] windowController] stopAnimation];
	    }
	}
    }

  // We check to see if there's any more task that needs to be fired
  [self _nextTask];
}


//
//
//
- (void) stopTask: (Task *) theTask
{
  Service *aService;
  NSArray *theKeys;
  int i, count;
  
  // We get the right object based on the out Task's object.
  theKeys = NSAllMapTableKeys(_table);
  aService = nil;

  count = [theKeys count];
  
  for (i = 0; i < count; i++)
    {
      if (NSMapGet(_table, [theKeys objectAtIndex: i]) == theTask)
	{
	  aService = [theKeys objectAtIndex: i];
	}
    }

  if (!aService)
    {
      return;
    }

  [theTask setRunning: NO];
  [aService cancelRequest];
}


//
// This method will return IIF the application is about to terminate.
//
- (void) run
{
  // We create our timer, it'll get added to the runloop in the NSDefaultRunLoopMode and NSEventTrackingRunLoopMode.
  _timer = [NSTimer timerWithTimeInterval: 5
		    target: self
		    selector: @selector(_tick)
		    userInfo: nil
		    repeats: YES];
  [[NSRunLoop currentRunLoop] addTimer: _timer  forMode: NSEventTrackingRunLoopMode];
  [[NSRunLoop currentRunLoop] addTimer: _timer  forMode: NSDefaultRunLoopMode];
  [[NSRunLoop currentRunLoop] addTimer: _timer  forMode: NSModalPanelRunLoopMode];
  RETAIN(_timer);
}


//
//
//
- (void) stop
{
  [_timer invalidate];
  RELEASE(_timer);
}


//
// Pantomime's delegate methods
//
- (void) authenticationCompleted: (NSNotification *) theNotification
{
  Task *aTask;
  id o;

  aTask = NSMapGet(_table, [theNotification object]);
  o = [theNotification object];

  if ([o isKindOfClass: SMTP_class])
    {
      ADD_CONSOLE_MESSAGE(_(@"SMTP - Authenticated! Sending the message..."));
      [o sendMessage];

      // We reupdate the message's size, since we now know its final value.
      aTask->totalSize = (float)[[o messageData] length]/(float)1024;;
    }
  else if ([o isKindOfClass: POP3Store_class])
    {
      POP3CacheManager *aCacheManager;		      
      NSString *aCacheFilename;
      
      // We get our POP3 cache
      aCacheFilename = [Utilities flattenPathFromString: [NSString stringWithFormat: @"%@ @ %@", [o username], [(Service*)o name]]
				  separator: @"/"];
      aCacheManager = [POP3CacheManager cacheWithPath: [NSString stringWithFormat: @"%@/POP3Cache_%@",
								 GNUMailUserLibraryPath(),
								 aCacheFilename]];
      [[o defaultFolder] setCacheManager: aCacheManager];
      
      ADD_CONSOLE_MESSAGE(_(@"POP3 - Authenticated! Transferring messages..."));
      [[o defaultFolder] prefetch];
    }
  else
    {
      NSString *aString;
      NSNumber *aNumber;

      ADD_CONSOLE_MESSAGE(_(@"IMAP - Authenticated!"));
      
      aString = [Utilities accountNameForServerName: [(Service *)o name]  username: [o username]];

      aNumber = [[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] 
		    objectForKey: aString] objectForKey: @"RECEIVE"] objectForKey: @"SHOW_WHICH_MAILBOXES"];
      
      if (aNumber && [aNumber intValue] == IMAP_SHOW_SUBSCRIBED_ONLY)
	{
	  [o subscribedFolderEnumerator];
	}
      else
	{
	  [o folderEnumerator];
	}
    }
}


//
//
//
#warning disconnect if imap auth failed and no task
- (void) authenticationFailed: (NSNotification *) theNotification
{
  NSString *aString, *aName;
  Task *aTask;
  id o;
  
  o = [theNotification object];
  aTask = NSMapGet(_table, [theNotification object]);

  // We get the right account name.
  if (aTask)
    {
      aName = [aTask key];
    }
  else
    {
      aName = [Utilities accountNameForServerName: [(Service *)o name]  username: [o username]];
    }


  if ([o isKindOfClass: POP3Store_class])
    {
      aString = _(@"POP3");
    }
  else if ([o isKindOfClass: IMAPStore_class])
    {
      aString = _(@"IMAP");
    }
  else
    {
      aString = _(@"SMTP");
    }

  NSRunAlertPanel(_(@"Error!"),
		  _(@"%@ authentication failed for account %@."),
		  _(@"OK"),
		  NULL,
		  NULL,
		  aString,
		  aName);
  
  // We now invalidate our password
  [[Utilities passwordCache] removeObjectForKey: [NSString stringWithFormat: @"%@ @ %@", [o username], [(Service*)o name]]];
  
  // We close the TCP connection
  [o close];
  
  // We leave the task in the queue if it's a SMTP related task.
  // Otherwise, we do nothing. We will do something else
  // in -connectionTerminated.
  if (aTask && [o isKindOfClass: SMTP_class])
    {
      [aTask setDate: [AUTORELEASE([[NSDate alloc] init]) addTimeInterval: 300]];
      [aTask setRunning: NO];
      [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
    }
  else if (!aTask && [o isKindOfClass: IMAPStore_class])
    {
      [[MailboxManagerController singleInstance] setStore: nil
						 name: [(Service*)o name]
						 username: [o username]];
    }
}


//
//
//
- (void) serviceInitialized: (NSNotification *) theNotification
{
  NSDictionary *allValues;
  NSString *aPassword;
  id o;

  o = [theNotification object];

  if ([o isKindOfClass: SMTP_class])
    {
      Task *aTask;
      
      aTask = NSMapGet(_table, o);

      ADD_CONSOLE_MESSAGE(_(@"SMTP - Connected to %@!"), [(SMTP *)o name]);

      // We must verify if we need to use SMTP authentifcation.
      allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
		     objectForKey: [aTask sendingKey]] objectForKey: @"SEND"];
     
      if ([allValues objectForKey: @"SMTP_AUTH"] &&
	  [[allValues objectForKey: @"SMTP_AUTH"] intValue] == NSOnState)
	{
	  ADD_CONSOLE_MESSAGE(_(@"SMTP - Authenticating to %@ using %@..."),
			      [allValues objectForKey: @"SMTP_HOST"],
			      [allValues objectForKey: @"SMTP_USERNAME"]);
	  
	  aPassword = [Utilities passwordForKey: [aTask sendingKey]  type: OTHER  prompt: YES];
	  
	  if (aPassword)
	    {
	      [o authenticate: [allValues objectForKey: @"SMTP_USERNAME"]
		 password: aPassword
		 mechanism: [allValues objectForKey: @"SMTP_AUTH_MECHANISM"]];
	    }
	  else
	    {
	      [self authenticationFailed: theNotification];
	    }
	}
      else
	{
	  // We do not need to use SMTP AUTH, let's send the message right away.
	  [o sendMessage];
	  
	  // We reupdate the message's size, since we now know its final value.
	  aTask->totalSize = (float)[[o messageData] length]/(float)1024;;
	}
    }
  else if ([o isKindOfClass: POP3Store_class] ||
	   [o isKindOfClass: IMAPStore_class])
    {
      NSString *aMechanism, *aString;
      
      aString = [Utilities accountNameForServerName: [(Service *)o name]  username: [o username]];

      allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] 
		     objectForKey: aString] objectForKey: @"RECEIVE"];
      aMechanism = nil;
      
      if ([o isKindOfClass: POP3Store_class])
	{
	  ADD_CONSOLE_MESSAGE(_(@"POP3 - Connected to %@!"), [(Service *)o name]);
	  
	  if ([allValues objectForKey: @"USEAPOP"])
	    {
	      aMechanism = ([[allValues objectForKey: @"USEAPOP"] intValue] == NSOnState ? @"APOP" : nil);
	    }
	}
      else
	{
	  ADD_CONSOLE_MESSAGE(_(@"IMAP - Connected to %@!"), [(Service *)o name]);
	  aMechanism = [allValues objectForKey: @"AUTH_MECHANISM"];
	  
	  if (aMechanism && [aMechanism isEqualToString: @"Password"])
	    {
	      aMechanism = nil;
	    }
	}
      
      aPassword = [Utilities passwordForKey: aString  type: IMAP  prompt: YES];
      
      if (aPassword)
	{
	  [o authenticate: [allValues objectForKey: @"USERNAME"]
	     password: aPassword
	     mechanism: aMechanism];
	}
      else
	{
	  [self authenticationFailed: theNotification];
	}
    }
}


//
//
//
- (void) serviceReconnected: (NSNotification *) theNotification
{

}


//
//
//
- (void) connectionEstablished: (NSNotification *) theNotification
{
  NSDictionary *allValues;
  id o;

  o = [theNotification object];

  if ([o isKindOfClass: SMTP_class])
    {
      Task *aTask;
      
      aTask = NSMapGet(_table, o);

      ADD_CONSOLE_MESSAGE(_(@"SMTP - Connected to %@!"), [(SMTP *)o name]);

      // We must verify if we need to use SMTP authentifcation.
      allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
		     objectForKey: [aTask sendingKey]] objectForKey: @"SEND"];

      // We check if we have to enable SSL
      if ([allValues objectForKey: @"USESECURECONNECTION"] &&
	  [[allValues objectForKey: @"USESECURECONNECTION"] intValue] == NSOnState)
	{
	  [(TCPConnection *)[o connection] startSSL];
	}
    }
  else if ([o isKindOfClass: POP3Store_class] ||
	   [o isKindOfClass: IMAPStore_class])
    {
      NSString *aString;
      
      aString = [Utilities accountNameForServerName: [(Service *)o name]  username: [o username]];

      allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] 
		     objectForKey: aString] objectForKey: @"RECEIVE"];

      // We check if we have to enable SSL
      if ([allValues objectForKey: @"USESECURECONNECTION"] &&
	  [[allValues objectForKey: @"USESECURECONNECTION"] intValue] == NSOnState)
	{
	  [(TCPConnection *)[o connection] startSSL];
	}
    }
}


//
//
//
- (void) connectionTerminated: (NSNotification *) theNotification
{
  Task *aTask;
  id o;

  o = [theNotification object];
  aTask = NSMapGet(_table, o);

  if (aTask)
    {
      if ([o isKindOfClass: POP3Store_class])
	{
	  if ( (![[NSUserDefaults standardUserDefaults] 
		   objectForKey: @"SHOW_NO_NEW_MESSAGES_PANEL"] ||
		[[NSUserDefaults standardUserDefaults]
		  integerForKey: @"SHOW_NO_NEW_MESSAGES_PANEL"] == NSOnState) &&
	       [aTask receivedMessagesCount] == 0)
	    {
	      NSRunAlertPanel(_(@"No New Messages..."),
			      _(@"There are no new messages on %@ @ %@."),
			      _(@"OK"),
			      NULL,
			      NULL,
			      [o username],
			      [(POP3Store *)o name]);
	    }
	  
	  ADD_CONSOLE_MESSAGE(_(@"No new messages on server %@"), [(POP3Store *)o name]);
	}
      
      NSMapRemove(_table, o);

      if ([o isKindOfClass: POP3Store_class] || [o isKindOfClass: IMAPStore_class])
	{
	  [self taskCompleted: aTask];
	}
    }

  AUTORELEASE(o);
}


//
//
//
- (void) connectionTimedOut: (NSNotification *) theNotification
{
  NSString *aString;
  Task *aTask;
  id o;

  o = [theNotification object];
  aTask = NSMapGet(_table, o);
  
  if ([o isKindOfClass: POP3Store_class])
    {
      aString = _(@"POP3");
    }
  else if ([o isKindOfClass: IMAPStore_class])
    {
      aString = _(@"IMAP");
    }
  else
    {
      aString = _(@"SMTP");
    }

  NSRunAlertPanel(_(@"Error!"),
		  _(@"Unable to communicate with the %@ server (%@)."),
		  _(@"OK"),
		  NULL,
		  NULL,
		  aString,
		  [(Service *)o name]);
  
  NSMapRemove(_table, o);

  if (aTask)
    {
      if ([o isKindOfClass: SMTP_class])
	{
	  [aTask setDate: [AUTORELEASE([[NSDate alloc] init]) addTimeInterval: 300]];
	  [aTask setRunning: NO];
	  [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
	}
      else
	{
	  [self taskCompleted: aTask];
	}
    }
  
  RELEASE(o);
}



//
//
//
- (void) folderExpungeCompleted: (NSNotification *) theNotification
{
  id o;

  o = [theNotification object];

  if ([o isKindOfClass: IMAPStore_class])
    {
      [[[[GNUMail lastMailWindowOnTop] windowController] folder] updateCache];
      [[[GNUMail lastMailWindowOnTop] windowController] tableViewShouldReloadData];
    }
  else if ([o isKindOfClass: LocalStore_class])
    {

    }
  else
    {
      [o close];
    }
}


//
//
//
- (void) folderExpungeFailed: (NSNotification *) theNotification
{
  
}


//
//
//
- (void) folderPrefetchCompleted: (NSNotification *) theNotification
{
  id o;

  o = [theNotification object];

  if ([o isKindOfClass: IMAPStore_class] || [o isKindOfClass: LocalStore_class])
    {
      [[[GNUMail lastMailWindowOnTop] windowController] updateDataView];
      [[[GNUMail lastMailWindowOnTop] windowController] stopAnimation];
      [[[[GNUMail lastMailWindowOnTop] windowController] dataView] scrollIfNeeded];
    }
  else if ([o isKindOfClass: POP3Store_class])
    {
      POP3CacheObject *aPOP3CacheObject;
      POP3Message *aMessage;
      POP3Folder *aFolder;
      NSString *aUID;
      Task *aTask;
      
      int i, count;
 
      aTask = NSMapGet(_table, o);
      aFolder = [o defaultFolder];
      count = [aFolder count];
   
      // We get all messages..
      for (i = 1; i <= count; i++)
	{
	  aMessage = (POP3Message *)[aFolder messageAtIndex: i-1];
	  
	  aUID = [aMessage UID];
	  
	  // Next we verify if we really need to transfer the message
	  aPOP3CacheObject = [[aFolder cacheManager] cacheObjectWithUID: aUID];
	  
	  // We do...
	  if (!aPOP3CacheObject)
	    {
	      [aFolder prefetchMessageAtIndex: i];
	      aTask->totalSize += (float)[aMessage size]/(float)1024;
	      aTask->totalMessagesCount += 1;
	    }
	}

      // If we haven't transferred any messages or if there's none
      // available on the server, we close the connection right away.
      if (aTask->totalMessagesCount == 0)
	{
	  [o close];
	}
    }
}


//
//
//
- (void) messagePrefetchCompleted: (NSNotification *) theNotification
{
  id o;

  o = [theNotification object];

  if ([o isKindOfClass: POP3Store_class])
    {
      Message *aMessage;
      Task *aTask;

      BOOL b;
      
      aTask = NSMapGet(_table, o);
      aMessage = [[theNotification userInfo] objectForKey: @"Message"];
      
      if (!aMessage)
	{
	  return;
	}
      
      [aTask setReceivedMessagesCount: [aTask receivedMessagesCount]+1];
      
      b = [self _matchFilterRuleFromRawSource: [aMessage rawSource]  task: aTask];
      
      // We cache the UID of the message we just got if we have been able to
      // save the message on disk. If not, we'll leave it on the server.
      if (b)
	{
	  POP3CacheObject *aCacheObject;
	  
	  aCacheObject = [[POP3CacheObject alloc] initWithUID: [(POP3Message *)aMessage UID]
						  date: [NSCalendarDate calendarDate]];
	  [[[[theNotification object] defaultFolder] cacheManager] addObject: aCacheObject];
	  RELEASE(aCacheObject);
	  
	  // We synchronize our POPCacheManager
	  [[[[theNotification object] defaultFolder] cacheManager] synchronize];

	  if (![(POP3Folder *)[aMessage folder] leaveOnServer])
	    {
	      [(POP3Folder *)[aMessage folder] deleteMessageAtIndex: [aMessage messageNumber]];
	    }
	}

      // If we worked on the last message in the folder, we are ready to close
      // the connection with the POP3 server.
      if ([aMessage messageNumber] == [[aMessage folder] count])
	{
	  [o close];
	}
    }
}


//
//
//
- (void) messageNotSent: (NSNotification *) theNotification
{
  NSString *aString;
  Task *aTask;

  if ([[theNotification object] isKindOfClass: SMTP_class])
    {
      aString = [NSString stringWithFormat: _(@"An error occured while sending the E-Mail. It might be a\nnetwork problem or an error in your sending preferences.\nLast response received from the server:\n\n(%d) %@\n\nTo save this E-Mail in the Drafts folder, open the Console window\nfrom the Windows menu and click on the \"Save in Drafts\" button."), [[theNotification object] lastResponseCode], AUTORELEASE([[NSString alloc] initWithData: [[theNotification object] lastResponse] encoding: NSASCIIStringEncoding])];
    }
  else
    {
      aString = _(@"An error occured while sending the E-Mail. The path to the\nmailer might be incorrect in your sending preferences.");
    }

  NSRunAlertPanel(_(@"Error!"),
		  aString,
		   _(@"OK"),
		  NULL,
		  NULL,
		  NULL);

#warning fixme remove o from the map
  aTask = NSMapGet(_table, [theNotification object]);

  // We leave the task in queue.
  [aTask setDate: [AUTORELEASE([[NSDate alloc] init]) addTimeInterval: 300]];
  [aTask setRunning: NO];
  [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
  [[theNotification object] close];
}


//
//
//
- (void) messageSent: (NSNotification *) theNotification
{
  Task *aTask;

  aTask = NSMapGet(_table, [theNotification object]);

  ADD_CONSOLE_MESSAGE(_(@"SMTP - Sent!"));

  if ([aTask subOp] != GNUMailRedirectMessage)
    {
      FilterManager *aFilterManager;
      URLName *theURLName;

      aFilterManager = (FilterManager *)[FilterManager singleInstance];
      
      theURLName = [aFilterManager matchedURLNameFromMessageAsRawSource: [[theNotification object] messageData]
				   type: TYPE_OUTGOING
				   key: [(Task *)NSMapGet(_table, [theNotification object]) key]
				   filter: nil];
      
      [[MailboxManagerController singleInstance] addMessage: [[theNotification object] messageData]
						 toFolder: theURLName];
    }

  if ([[theNotification object] isKindOfClass: SMTP_class])
    {
      [[theNotification object] close];
    }

  [self taskCompleted: aTask];
}


//
//
//
- (void) service: (Service *) theService 
    receivedData: (NSData *) theData
{
  Task *aTask;

  aTask = NSMapGet(_table, theService);
  
  if (aTask)
    {
      aTask->currentSize += (float)([theData length]/(float)1024);
      [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
    }
}


//
//
//
- (void) service: (Service *) theService
	sentData: (NSData *) theData
{
  Task *aTask;

  aTask = NSMapGet(_table, theService);

  if (aTask)
    {
      aTask->currentSize += (float)([theData length]/(float)1024);
      [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
    }
}


//
//
//
- (void) requestCancelled: (NSNotification *) theNotification
{
  Task *aTask;
  id o;

  o = [theNotification object];
  aTask = NSMapGet(_table, o);

  if ([o isKindOfClass: SMTP_class])
    {
      [aTask setDate: [AUTORELEASE([[NSDate alloc] init]) addTimeInterval: 300]];
      [aTask setRunning: NO];
      aTask->currentSize = 0;
    }
  else
    {
      [self taskCompleted: aTask];
    }

  [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
}


//
//
//
- (void) folderListCompleted: (NSNotification *) theNotification
{
  Task *aTask;
  id o;
  
  o = [theNotification object];
  aTask = NSMapGet(_table, o);
  
  if (aTask)
    {
      NSArray *subscribedFolders;
      subscribedFolders = [NSArray arrayWithArray: [[[theNotification userInfo] objectForKey: @"NSEnumerator"] allObjects]];
#warning this is a workaround
      aTask->totalMessagesCount = [subscribedFolders count];
      [o folderStatus: subscribedFolders];
      [o close];
    }
  else
    {
      [[MailboxManagerController singleInstance] reloadFoldersForStore: o
						 folders: [[theNotification userInfo] objectForKey: @"NSEnumerator"]];
      [Utilities restoreOpenFoldersForStore: o];
    }
}


//
//
//
- (void) folderListSubscribedCompleted: (NSNotification *) theNotification
{
  [self folderListCompleted: theNotification];
}


//
//
//
- (void) folderStatusCompleted: (NSNotification *) theNotification
{
  FolderInformation *aFolderInformation;
  NSDictionary *aDictionary;
  NSString *aFolderName;
  
  Task *aTask;
  id o;
  
  o = [theNotification object];
  aTask = NSMapGet(_table, o);

#warning this is a workaround
  aTask->receivedMessagesCount += 1;

  aFolderInformation = [[theNotification userInfo] objectForKey: @"FolderInformation"];
  aFolderName = [[theNotification userInfo] objectForKey: @"FolderName"];

  aDictionary = [NSDictionary dictionaryWithObjectsAndKeys: aFolderInformation, @"FOLDER_INFORMATION",
			      aFolderName, @"FOLDER_NAME",
			      [(Service *)o name], @"STORE_NAME",
			      [o username], @"USERNAME", 
			      [o folderSeparator], @"FOLDER_SEPARATOR", nil];
  
  [[MailboxManagerController singleInstance] updateFolderInformation: aDictionary];
}


//
//
//
- (void) folderOpenCompleted: (NSNotification *) theNotification
{
  id aFolder;

  aFolder = [[theNotification userInfo] objectForKey: @"Folder"];
  [aFolder prefetch];
}


//
//
//
- (void) folderOpenFailed: (NSNotification *) theNotification
{
  NSArray *allWindows;
  id aFolder;

  aFolder = [[theNotification userInfo] objectForKey: @"Folder"];
  allWindows = [GNUMail allMailWindows];
  
  if (allWindows)
    {
      id aWindow;
      int i;

      for (i = 0; i < [allWindows count]; i++)
	{
	  aWindow = [allWindows objectAtIndex: i];
	  if ([(MailWindowController *)[aWindow windowController] folder] == aFolder)
	    {
	      NSRunAlertPanel(_(@"Error!"),
			      _(@"Unable to open mailbox %@ on %@."),
			      _(@"OK"),
			      NULL,
			      NULL,
			      [(Folder *)aFolder name],
			      [(Service *)[theNotification object] name]);
	      [(MailWindowController *)[aWindow windowController] setFolder: nil];
	      break;
	    }
	}
    }
}


//
// Class methods
//
+ (id) singleInstance
{
  if (!singleInstance)
    {
      singleInstance = [[TaskManager alloc] init];
    }

  return singleInstance;
}

@end



//
// TaskManager private interface
//
@implementation TaskManager (Private)

- (void) _executeActionUsingFilter: (Filter *) theFilter
                           message: (NSData *) theMessage
                              task: (Task *) theTask
{
  switch ([theFilter actionEMailOperation])
    {
    case BOUNCE:
    case FORWARD:
    case REPLY:
      NSLog(@"Unimplemented action - ignoring.");
      break;

    default:
      NSLog(@"Unknown action - ignoring.");
    }
}


//
//
//
- (BOOL) _matchFilterRuleFromRawSource: (NSData *) theRawSource
				  task: (Task *) theTask
{
  FilterManager *aFilterManager;
  NSString *aFolderName;
  URLName *theURLName;
  Filter *aFilter;
  
  aFilterManager = (FilterManager *)[FilterManager singleInstance];
  aFilter = [aFilterManager matchedFilterForMessageAsRawSource: theRawSource  type: TYPE_INCOMING];
  aFolderName = nil;

  if (aFilter && [aFilter action] == BOUNCE_OR_FORWARD_OR_REPLY)
    {
      [self _executeActionUsingFilter: aFilter
	    message: theRawSource
	    task: theTask];
    }
  else if (aFilter && [aFilter action] == PLAY_SOUND)
    {
      if ([[NSFileManager defaultManager] fileExistsAtPath: [aFilter pathToSound]])
	{
	  NSSound *aSound;
	  
	  aSound = [[NSSound alloc] initWithContentsOfFile: [aFilter pathToSound]  byReference: YES];
	  [aSound play];
	  RELEASE(aSound);
	}
    }
  
  // Even if we bounced (or forward or replied to the) message, or played a sound
  // when receiving it, we append it to our folder.
  theURLName = [aFilterManager matchedURLNameFromMessageAsRawSource: theRawSource
			       type: TYPE_INCOMING
			       key: [theTask key]
			       filter: aFilter];
  
  if ([theTask origin] == ORIGIN_USER)
    { 
      // We verify if our task owner is a MailWindowController. If it's the case,
      // we must check if we really need to add the the folder to our list of
      // filtered message folders.
      if ([theTask owner] && 
	  [[theTask owner] respondsToSelector: @selector(dataView)] &&
	  [Utilities URLWithString: [theURLName stringValue]
		     matchFolder: [[theTask owner] folder]])
	{
	  // Same folder, we skip it.
	  goto done;
	}
      
      if ([[theURLName protocol] caseInsensitiveCompare: @"LOCAL"] == NSOrderedSame)
	{
	  aFolderName = [NSString stringWithFormat: _(@"Local - %@"), [theURLName foldername]];
	}
      else
	{
	  aFolderName = [NSString stringWithFormat: _(@"IMAP %@@%@ - %@"), [theURLName username],
				  [theURLName host], [theURLName foldername]];
	}
    }
  
 done:
  if (![[MailboxManagerController singleInstance] addMessage: theRawSource  toFolder: theURLName])
    {
      return NO;
    }
  else if (aFolderName)
    {
      [theTask setFilteredMessagesCount: [theTask filteredMessagesCount]+1];
      
      if (![[theTask filteredMessagesFolders] containsObject: aFolderName])
	{
	  [[theTask filteredMessagesFolders] addObject: aFolderName];
	}
    }

  // We set the application icon to GNUMail_Full.tiff
  [NSApp setApplicationIconImage: [NSImage imageNamed: @"GNUMail_Full.tiff"]];

  return YES;
}


//
//
//
- (void) _nextTask
{ 
  Task *aTask;
  int i;

  aTask = nil;

  //
  // This method is the ONLY one that assigns the current running task.
  // If there's already one set, it returns nil.
  //   
  // We search for a task that should be run immediately
  for (i = ([_tasks count] - 1); i >= 0; i--)
    {
      aTask = [_tasks objectAtIndex: i];

      if (![aTask running])
	{
	  // We found one, let's use it.
	  if ([aTask immediate])
	    {
	      [aTask setRunning: YES];
	      [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
	      break;
	    }
	}
      
      aTask = nil;
    }  
  
  // We haven't found an immediate task. We loop from the oldest task
  // to the newest one to find a task that -date meets the requirements of
  // the current date/time.
  if (!aTask)
    {
      NSDate *aDate;

      aDate = [NSDate date];
      
      for (i = ([_tasks count] - 1); i >= 0; i--)
	{
	  aTask = [_tasks objectAtIndex: i];
	  
	  if (![aTask running])
	    {
	      if ([[aTask date] compare: aDate] == NSOrderedAscending)
		{
		  [aTask setRunning: YES];
		  [[[ConsoleWindowController singleInstance] tasksTableView] setNeedsDisplay: YES];
		  break;
		}
	    }

	  aTask = nil;
	}
    }

  if (aTask)
    {
      NSAutoreleasePool *pool;
      
      pool = [[NSAutoreleasePool alloc] init];

      switch ([aTask op])
	{
	case SEND_SENDMAIL:
	  [self _sendUsingSendmailForTask: aTask];
	  break;
	  
	case SEND_SMTP:
	  [self _sendUsingSMTPForTask: aTask];
	  break;

	case RECEIVE_IMAP:
	  [self _receiveUsingIMAPForTask: aTask];
	  break;

	case RECEIVE_POP3:
	  [self _receiveUsingPOP3ForTask: aTask];
	  break;

	case RECEIVE_UNIX:
	  [self _receiveUsingUNIXForTask: aTask];
	  break;

	default:
	  NSDebugLog(@"Unknown task type. Ignoring and keeping in the queue.");
	}

      RELEASE(pool);
    }
}


//
// We must get all body parts (download using body.peek[x]<from.increment>)
//
- (void) _receiveUsingIMAPForTask: (Task *) theTask
{
  NSDictionary *allValues;
  IMAPStore *aStore;
   
  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
		 objectForKey: [theTask key]] objectForKey: @"RECEIVE"];
  
  aStore = [[IMAPStore alloc] initWithName: [allValues objectForKey: @"SERVERNAME"]
			      port: [[allValues objectForKey: @"PORT"] intValue]];
  [aStore addRunLoopMode: NSEventTrackingRunLoopMode];
  [aStore addRunLoopMode: NSModalPanelRunLoopMode];
  [aStore setUsername: [allValues objectForKey: @"USERNAME"]];

  NSMapInsert(_table, aStore, theTask);
  [aStore setDelegate: self];
  [aStore connectInBackgroundAndNotify];
}


//
//
//
- (void) _receiveUsingPOP3ForTask: (Task *) theTask
{
  NSDictionary *allValues;
  POP3Store *aStore;

  // We get the values associated with the key
  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] 
		 objectForKey: [theTask key]] objectForKey: @"RECEIVE"];

  aStore = [[POP3Store alloc] initWithName: [allValues objectForKey: @"SERVERNAME"]
			      port: [[allValues objectForKey: @"PORT"] intValue]];
  [aStore addRunLoopMode: NSEventTrackingRunLoopMode];
  [aStore addRunLoopMode: NSModalPanelRunLoopMode];
  [aStore setUsername: [allValues objectForKey: @"USERNAME"]]; 

  NSMapInsert(_table, aStore, theTask);
  [aStore setDelegate: self];

  // We set some attributes to the POP3Folder
  [[aStore defaultFolder] setLeaveOnServer: ([[allValues objectForKey: @"LEAVEONSERVER"] intValue] == NSOnState ? YES : NO)];
  [[aStore defaultFolder] setRetainPeriod: [[allValues objectForKey: @"RETAINPERIOD"] intValue]];
  [aStore connectInBackgroundAndNotify];
}


//
//
//
- (void) _receiveUsingUNIXForTask: (Task *) theTask
{
  LocalMailDelivery *aLocalMailDelivery;
  NSDictionary *allValues;
  NSAutoreleasePool *pool;
  NSArray *anArray;
  int i;

  // We get our values associated with the key
  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] 
		 objectForKey: [theTask key]] objectForKey: @"RECEIVE"];

  aLocalMailDelivery = [[LocalMailDelivery alloc] initWithPathToSpool:
						    [allValues objectForKey: @"MAILSPOOLFILE"]];
  
  anArray = [aLocalMailDelivery messagesFromMailspoolFile];  
  pool = nil;
  
  for (i = 0; i < [anArray count]; i++)
    {
      if ( (i % 3) == 0 )
	{
	  TEST_RELEASE(pool);
	  pool = [[NSAutoreleasePool alloc] init];
	}
      
      [self _matchFilterRuleFromRawSource: [anArray objectAtIndex: i]
	    task: theTask];
    }
  
  TEST_RELEASE(pool);
  RELEASE(aLocalMailDelivery);

  [self taskCompleted: theTask];
}


//
//
//
- (void) _sendUsingSendmailForTask: (Task *) theTask
{
  NSDictionary *allValues;
  Sendmail *aSendmail;

  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
		 objectForKey: [theTask sendingKey]] objectForKey: @"SEND"];

  aSendmail = [[Sendmail alloc] initWithPath: [allValues objectForKey: @"MAILER_PATH"]]; 

  NSMapInsert(_table, aSendmail, theTask);

  [aSendmail setDelegate: self];
  
  if ([[theTask message] isKindOfClass: [NSData class]])
    {
      [aSendmail setMessageData: [theTask message]];
    }
  else
    {
      [aSendmail setMessage: [theTask message]];
    }

  [aSendmail sendMessage];
}


//
//
//
- (void) _sendUsingSMTPForTask: (Task *) theTask
{
  NSDictionary *allValues;
  NSNumber *portValue;
  SMTP *aSMTP;

  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"]
		 objectForKey: [theTask sendingKey]] objectForKey: @"SEND"];

  portValue = [allValues objectForKey: @"SMTP_PORT"];
  
  if (!portValue)
    {
      portValue = [NSNumber numberWithInt: 25];
    }

  aSMTP = [[SMTP alloc] initWithName: [allValues objectForKey: @"SMTP_HOST"]
			port: [portValue intValue]];
  [aSMTP addRunLoopMode: NSEventTrackingRunLoopMode];
  [aSMTP addRunLoopMode: NSModalPanelRunLoopMode];
  NSMapInsert(_table, aSMTP, theTask);

  [aSMTP setDelegate: self];

  if ([[theTask message] isKindOfClass: [NSData class]])
    {
      [aSMTP setMessageData: [theTask message]];
    }
  else
    {
      [aSMTP setMessage: [theTask message]];
    }
  
  ADD_CONSOLE_MESSAGE(_(@"SMTP - connecting to %@..."), [allValues objectForKey: @"SMTP_HOST"]);
  [aSMTP connectInBackgroundAndNotify];
}


//
// This method checks for new mail on the specified account name. 
//
// If the account type is POP3 or UNIX, this method adds a new task to 
// immediately check for new mails on this account.
//
// If the account type is IMAP, it NOOPs the folder or the store if
// and only if it is open.
//
- (void) _checkMailForAccount: (NSString *) theAccountName
		       origin: (int) theOrigin
			owner: (id) theOwner
{
  NSDictionary *allValues;
  Task *aTask;
  
  int op, subOp;
 
  // If the account is disabled or is set to never check mail we do nothing.
  if (![[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] objectForKey: theAccountName]
	  objectForKey: @"ENABLED"] boolValue] ||
      [[[[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] objectForKey: theAccountName]
	  objectForKey: @"RECEIVE"] objectForKey: @"RETRIEVEMETHOD"] intValue] == NEVER)
    {
      return;
    }

  allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] 
		 objectForKey: theAccountName] objectForKey: @"RECEIVE"];
  subOp = 0;
  
  if (![allValues objectForKey: @"SERVERTYPE"] || [[allValues objectForKey: @"SERVERTYPE"] intValue] == POP3)
    {
      op = RECEIVE_POP3;
    }
  else if ([[allValues objectForKey: @"SERVERTYPE"] intValue] == IMAP)
    {
      IMAPStore *aStore;
      
      aStore = [[MailboxManagerController singleInstance] storeForName: [allValues objectForKey: @"SERVERNAME"]
							  username: [allValues objectForKey: @"USERNAME"]];
      
      // We noop the store, if one is initialized.
      if (aStore)
	{  
	  ADD_CONSOLE_MESSAGE(_(@"NOOPing IMAP server %@."), [aStore name]);
	  [aStore noop];
	}
      else
	{
	  // The IMAP store is closed, no need to check for new mail on that one.
	  return;
	}
      
      op = RECEIVE_IMAP;
      subOp = IMAP_STATUS;
    }
  else
    {
      op = RECEIVE_UNIX;
    }
  
  aTask = [[Task alloc] init];
  [aTask setOp: op];
  [aTask setSubOp: subOp];
  [aTask setMessage: nil];
  [aTask setKey: theAccountName];
  [aTask setImmediate: YES];
  [aTask setOrigin: theOrigin];
  [aTask setOwner: theOwner];
  [self addTask: aTask];
  RELEASE(aTask);
}


//
//
//
- (void) _tick
{
  _counter += 5;
  
  // Every single minute, we invoke _tick:
  if (_counter%12 == 0)
    {
      [self _tick_internal];
    }
}


//
//
//
- (void) _tick_internal
{
  NSEnumerator *theEnumerator, *foldersEnumerator;
  id aController, aStore, aFolder;
  NSMutableArray *allFolders;
  NSArray *allWindows;

  NSDictionary *allValues;
  NSString *aKey;
  
  aController = [GNUMail lastMailWindowOnTop];

  if (aController)
    {
      aController = [[GNUMail lastMailWindowOnTop] windowController];

      if ([aController isKindOfClass: [MessageViewWindowController class]])
	{
	  aController = [(MessageViewWindowController *)aController mailWindowController];
	}
    }
 
  theEnumerator = [[Utilities allEnabledAccounts] keyEnumerator];

  while ((aKey = [theEnumerator nextObject]))
    {     
      allValues = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] 
		     objectForKey: aKey] objectForKey: @"RECEIVE"];
      
      // If we check for new mails from this account automatically AND
      // If our counter modulo the retreive interval is zero, we check for new mails.
      if ([[allValues objectForKey: @"RETRIEVEMETHOD"] intValue] == AUTOMATICALLY &&
	  ((_counter/12) % [[allValues objectForKey: @"RETRIEVEMINUTES"] intValue] ) == 0 )
	{
	  [self _checkMailForAccount: aKey
		origin: ORIGIN_TIMER
		owner: aController];
	}
    }

  allFolders = [[NSMutableArray alloc] init];

  //
  // We now loop in all opened MailWindow:s and all messages of them to verify if we got
  // any messages that are expired, and not selected.   
  //
  allWindows = [GNUMail allMailWindows];
  
  if (allWindows)
    {
      NSArray *allMessages;
      Folder *aFolder;

      NSCalendarDate *date;
      NSDate *aDate;

      id aWindow, aMessage, aSelectedMessage;
      int i, j, count, seconds;
      
      date = (NSCalendarDate *)[NSCalendarDate date];

      for (i = 0; i < [allWindows count]; i++)
	{
	  aWindow = [allWindows objectAtIndex: i];
	  aFolder = [(MailWindowController *)[aWindow windowController] folder];

	  // If the window has no selected folder, skip it.
	  if (!aFolder) continue;
	    
	  [allFolders addObject: aFolder];

	  aSelectedMessage = [(MailWindowController *)[aWindow windowController] selectedMessage];

	  allMessages = [aFolder allMessages];
	  count = [aFolder count];
	  
	  for (j = 0; j < count; j++)
	    {
	      aMessage = [allMessages objectAtIndex: j];
	      
	      // If this message is selected, skip it
	      if (aMessage == aSelectedMessage)
		{
		  continue;
		}

	      aDate = [aMessage propertyForKey: @"EXPIRE_DATE"];
	      
	      // If this property doesn't exist, that means it was never initialzed, so skip it
	      if (!aDate) 
		{
		  continue;
		}
	      
	      [date years: NULL
		    months: NULL
		    days: NULL
		    hours: NULL
		    minutes: NULL
		    seconds: &seconds
		    sinceDate: aDate];
	      
	      // If the message has expired, we release its content and we set it as uninitialzed
	      if (seconds > RETAIN_PERIOD)
		{
		  [aMessage setInitialized: NO];
		  [aMessage setRawSource: nil];
		  [(Message *)aMessage setProperty: nil  forKey: @"EXPIRE_DATE"];
		}
	    }
	}
    }

  //
  // We verify if we must expire (ie., close) some folders that were likely left into
  // an open state after a DnD operation.
  //
  theEnumerator = [[[MailboxManagerController singleInstance] allStores] objectEnumerator];

  while ((aStore = [theEnumerator nextObject]))
    {
      foldersEnumerator = [aStore openFoldersEnumerator];

      while ((aFolder = [foldersEnumerator nextObject]))
	{
	  if (![allFolders containsObject: aFolder])
	    {
	      NSDate *aDate;
	      int seconds;

	      aDate = [aFolder propertyForKey: @"EXPIRE_DATE"];

	      if (!aDate)
		{
		  continue;
		}

	      [(NSCalendarDate *)[NSCalendarDate date] years: NULL
				 months: NULL
				 days: NULL
				 hours: NULL
				 minutes: NULL
				 seconds: &seconds
				 sinceDate: aDate];
	            
	      if (seconds > RETAIN_PERIOD)
		{
		  [aFolder close];
		}
	    }
	}
    }

  RELEASE(allFolders);
}

@end
