/*
**  ThreadArcsCell.m
**
**  Copyright (c) 2004
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ThreadArcsCell.h"

#include "Constants.h"
#include "MailHeaderCell.h"
#include "MailWindowController.h"
#include "NSBezierPath+Extensions.h"
#include "NSFont+Extensions.h"

#include <Pantomime/Container.h>
#include <Pantomime/Folder.h>
#include <Pantomime/Message.h>

//
// Private interface
//
@interface ThreadArcsCell (Private)

- (void) _drawArcsInFrame: (NSRect) theFrame
		   inView: (NSView *) theView;

@end


//
//
//
@implementation ThreadArcsCell

- (id) init
{
  self = [super init];

  _string = [[NSMutableAttributedString alloc] initWithString: _(@"Thread")];
  [_string setAttributes: [NSDictionary dictionaryWithObjectsAndKeys: [NSFont headerNameFont], NSFontAttributeName,
					[NSColor colorWithCalibratedRed: 0.686  green: 0.686  blue: 0.686  alpha: 1.0],
					NSForegroundColorAttributeName, nil]
	   range: NSMakeRange(0, [_string length])];
  _color = RETAIN([NSColor colorWithCalibratedRed: 0.898
			   green: 0.988
			   blue: 0.937
			   alpha: 1.0]);

  _rect_table = NSCreateMapTable(NSObjectMapKeyCallBacks, NSObjectMapValueCallBacks, 16);

  return self;
}


//
//
//
- (void) dealloc
{
  RELEASE(_string);
  RELEASE(_color);
  NSFreeMapTable(_rect_table);
  [super dealloc];
}


//
//
//
- (BOOL) trackMouse: (NSEvent *) theEvent
	     inRect: (NSRect) cellFrame
	     ofView: (NSView *) aTextView
   atCharacterIndex: (unsigned) charIndex
       untilMouseUp: (BOOL) flag
{
  NSArray *allKeys;
  NSValue *aValue;

  NSPoint aPoint;
  NSRect aRect;
  int i;

  aPoint = [aTextView convertPoint: [theEvent locationInWindow]  fromView: nil];
  allKeys = NSAllMapTableKeys(_rect_table);
  
  for (i = 0; i < [allKeys count]; i++)
    {
      aValue = [allKeys objectAtIndex: i];
      aRect = [aValue rectValue];
      
      if (NSMouseInRect(aPoint, aRect, YES))
	{
	  Message *aMessage;
	  int row;
	  
	  aMessage = NSMapGet(_rect_table, aValue);
	  row = [[[(NSTextView *)aTextView delegate] allMessages] indexOfObject: aMessage];
	  [[[(NSTextView *)aTextView delegate] dataView] selectRow: row  byExtendingSelection: NO];
	}
    }

  return YES;
}

//
//
//
- (void) drawWithFrame: (NSRect) theFrame
		inView: (NSView *) theView
{
  NSBezierPath *aRoundRect;
  id aCell;

  int x;

  aCell = [[[theView window] windowController] mailHeaderCell];
  theFrame.size.height = [aCell height];
  theFrame.size.width = 140;
  theFrame.origin.x -= 3;
  
  // FIXME
  //#ifdef MACOSX
  theFrame.origin.y = 5;
  //#endif

  [_color set];

  aRoundRect = [NSBezierPath bezierPath];
  [aRoundRect appendBezierPathWithRoundedRectangle: theFrame
                                        withRadius: 8.0];
  [aRoundRect fill];

  x = (140-[_string size].width)/2 + theFrame.origin.x;
  [_string drawAtPoint: NSMakePoint(x, 10)];

  [self _drawArcsInFrame: theFrame  inView: theView];
}

@end


//
//
//
@implementation ThreadArcsCell (Private)

- (void) _drawArcsInFrame: (NSRect) theFrame
		   inView: (NSView *) theView
{
  NSMutableArray *allMessages;
  NSEnumerator *theEnumerator;
  Container *aContainer;
  Message *aMessage;

  NSBezierPath *aBezierPath;
  NSRect aRect;
  int i, j, origin, count;
  BOOL flip;

  NSResetMapTable(_rect_table);

  aMessage = [[[theView window] windowController] selectedMessage];
  
  //NSLog(@"Subject = %@", [aMessage subject]);

  //
  // Find the root container
  //
  aContainer = [aMessage propertyForKey: @"Container"];

  if (!aContainer)
    {
      //NSLog(@"FOUND NO CONTAINER");
      return;
    }

  while (aContainer->parent) aContainer = aContainer->parent;
  
  if (!aContainer->message)
    {
      //NSLog(@"NO ASSOCIATED MESSAGE");
      return;
    }

  //
  // We now get all children of the root container. We need them
  // in order to fully draw the thread arcs.
  // 
  allMessages = [[NSMutableArray alloc] init];
  [allMessages addObject: aContainer->message];
  theEnumerator = [aContainer childrenEnumerator];

  while ((aContainer = [theEnumerator nextObject]))
    {
      aMessage = aContainer->message;
      //NSLog(@"child = %@", [aMessage subject]);
      [allMessages addObject: aMessage];
    }

  //
  // We sort the messages by date, critical for thread arcs to be useful.
  //
  [allMessages sortUsingSelector: @selector(compareAccordingToDate:)];
  count = [allMessages count];
  origin = theFrame.origin.x+10;

  //
  // We draw the dots
  // 
  for (i = 0; i < count; i++)
    {
      aMessage = [allMessages objectAtIndex: i];
      aRect = NSMakeRect(origin+i*10, 80, 5, 5);

      NSMapInsert(_rect_table, (void *)[NSValue valueWithRect: aRect], (void *)aMessage);

      if (aMessage == [[[theView window] windowController] selectedMessage])
	{
	  [[NSColor blueColor] set];
	  [[NSBezierPath bezierPathWithOvalInRect: aRect] stroke];
	}
      else
	{
	  [[NSColor lightGrayColor] set];
	  [[NSBezierPath bezierPathWithOvalInRect: aRect] fill];
	}
    }

  
  //
  // We draw the arcs
  //
  for (i = 0; i < count; i++)
    {
      aMessage = [allMessages objectAtIndex: i];     
      theEnumerator = [[aMessage propertyForKey: @"Container"] childrenEnumerator];
 
      while ((aContainer = [theEnumerator nextObject]))
	{
	  // We do NOT consider the sub-children
	  if (aContainer->parent != [aMessage propertyForKey: @"Container"])
	    {
	      //NSLog(@"SKIPPING SUB-CHILDREN");
	      continue;
	    }

	  j = [allMessages indexOfObject: aContainer->message]; 
	  flip = (j%2 == 0 ? YES : NO);

	  //NSLog(@"i === %d   j ===== %d   flip = %d", i, j, flip);
	  //if (flip) NSLog(@"FLIP!");

	  //
	  // If we are drawing arcs for the selected message OR
	  // if we are drawing arcs TO the selected message
	  //
	  if (aMessage == [[[theView window] windowController] selectedMessage] ||
	      aContainer->message == [[[theView window] windowController] selectedMessage])
	    {
	      [[NSColor blueColor] set];
	    }
	  else
	    {
	      [[NSColor lightGrayColor] set];
	    }

	  aBezierPath = [NSBezierPath bezierPath];
	  [aBezierPath appendBezierPathWithArcWithCenter: NSMakePoint(origin+7.5+i*10+(j-i-1)*5, (flip?85:81))
		       radius: 5+(j-i-1)*5
		       startAngle: (flip?0:180)
		       endAngle: (flip?180:0)];
	  [aBezierPath stroke];
	}
    }

  RELEASE(allMessages);
}

@end
